#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import

import unittest

from openfermionprojectq import Graph, Node


class GraphTest(unittest.TestCase):

    def test_two_node_graph(self):
        """Build a graph with two nodes and one edge"""
        two_node = Graph()
        two_node.add_node()
        two_node.add_node()
        two_node.add_edge(0, 1)

        # Check to see if it built normally
        self.assertEqual(two_node.neighbors,
                         [{1}, {0}])
        self.assertEqual(two_node.node_uids,
                         [0, 1])

        self.assertEqual(two_node.uid_to_index[0],
                         0)
        self.assertEqual(two_node.uid_to_index[1],
                         1)
        self.assertEqual(two_node.node_count(), 2)

        # Now remove an edge, then a node, see if it
        two_node.remove_edge(0, 1)
        self.assertEqual(two_node.neighbors,
                         [set([]), set([])])
        two_node.remove_node(0)
        self.assertEqual(two_node.node_uids, [1])

        # Add a node and edge back, verify indexing
        two_node.add_node()
        two_node.add_edge(0, 1)
        self.assertEqual(two_node.neighbors,
                         [{2}, {1}])
        self.assertEqual(two_node.uid_to_index[1], 0)
        self.assertEqual(two_node.uid_to_index[2], 1)
        self.assertEqual(two_node.node_count(), 2)

        self.assertEqual(two_node.get_neighbors(0), [1])
        self.assertEqual(two_node.get_neighbors(1), [0])

        self.assertTrue(two_node.is_adjacent(0, 1))

        # Check path between nodes
        self.assertEqual(two_node.shortest_path(0, 1), [0, 1])
        self.assertEqual(two_node.shortest_path(1, 0), [1, 0])
        self.assertEqual(two_node.shortest_path(0, 0), [0])

        # Remove just the node and check neighbors
        two_node.remove_node(1)
        self.assertEqual(two_node.neighbors,
                         [set([])])

    def test_ring(self):
        """Build a ring of 8 nodes and test path finding"""
        eight_node = Graph()
        for i in range(8):
            eight_node.add_node(Node(value=i))
        for i in range(8):
            eight_node.add_edge(i, (i + 1) % 8)
        self.assertEqual(eight_node.neighbors,
                         [{1, 7}, {0, 2}, {1, 3}, {2, 4},
                          {3, 5}, {4, 6}, {5, 7}, {0, 6}])

        self.assertEqual(eight_node.shortest_path(0, 4),
                         [0, 7, 6, 5, 4])
        self.assertEqual(eight_node.shortest_path(0, 6),
                         [0, 7, 6])
        self.assertTrue(eight_node.is_adjacent(0, 7))

        # Look for node with value 6 and value not present
        found_index = eight_node.find_index(6)
        self.assertEqual(found_index, 6)
        found_index = eight_node.find_index(10)
        self.assertIsNone(found_index)

        # Make a hole in the ring and check new distance
        eight_node.remove_node(7)
        self.assertTrue(eight_node.is_adjacent(0, 1))
        self.assertFalse(eight_node.is_adjacent(0, 6))
        self.assertEqual(eight_node.shortest_path(0, 6),
                         [0, 1, 2, 3, 4, 5, 6])



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import

import queue


class Node:
    """Graph node

    These graph nodes may be used to store data or other attributes within
    the Graph structure.
    """
    def __init__(self, value=None):
        """Build a graph node initialized with generic value"""
        self.value = value


class Graph:
    """An undirected graph of nodes of arbitrary connectivity

    A generic graph class for undirected graphs that holds Nodes and
    edges that connect them.

    Attributes:
        nodes(list): A list of Node objects containing the nodes of the graph
        node_uids(list): A list of unique IDs (uids) for each node
        uid_to_index(dict): A dictionary that maps UIDs currently present
            in the graph to their node index
        neigbors(list of sets):  A list of sets that enumerate the neighbors
            of each node.  For example the neighbors of node i are the set
            neighbors[i]
        next_uid(int):  The next unique ID to be assigned to a node on addition
    """

    def __init__(self):
        """Set up an empty graph with no nodes"""
        self.nodes = []
        self.node_uids = []
        self.uid_to_index = {}
        self.neighbors = []
        self.next_uid = 0

    def add_node(self, node=Node()):
        """Add a Node to the graph.

        Args:
            node(Node): A Node object to be added to the graph

        Returns(int): The unique identified that was given to the node
        """
        self.nodes += [node]
        self.node_uids += [self.next_uid]
        self.neighbors += [set()]
        self.uid_to_index[self.next_uid] = self.node_count() - 1
        self.next_uid += 1

    def remove_node(self, node_id):
        """Remove a graph node

        This removes a node from the graph and leverages the unique ID
        system used internally to avoid having to modify the edges for all
        nodes in the graph.

        Args:
            node_id(int): Index of the node to be removed.
        """
        if node_id >= self.node_count():
            raise IndexError("Node ID out of range for graph")
        node_uid = self.node_uids[node_id]
        for i in range(self.node_count()):
            if node_uid in self.neighbors[i]:
                self.neighbors[i].remove(node_uid)
        for i in range(node_id + 1, self.node_count()):
            self.uid_to_index[self.node_uids[i]] -= 1

        del self.uid_to_index[node_uid]
        del self.nodes[node_id]
        del self.neighbors[node_id]
        del self.node_uids[node_id]

    def node_count(self):
        """Number of nodes in the graph

        Returns(int): Number of nodes currently in the graph
        """
        return len(self.nodes)

    def add_edge(self, node_id1, node_id2):
        """Add an edge between node 1 and node2

        Args:
            node_id1(int): Index of first node on edge
            node_id2(int): Index of second node on edge
        """
        if node_id1 >= self.node_count() or node_id2 >= self.node_count():
            raise IndexError("Node ID out of range for graph.")
        if node_id1 == node_id2:
            raise IndexError("Error, self-loops not supported.")

        self.neighbors[node_id1].add(self.node_uids[node_id2])
        self.neighbors[node_id2].add(self.node_uids[node_id1])

    def remove_edge(self, node_id1, node_id2):
        """Remove an edge between node1 and node2

        Args:
            node_id1(int): Index of first node on edge
            node_id2(int): Index of second node on edge
        """
        if node_id1 >= self.node_count() or node_id2 >= self.node_count():
            raise IndexError("Node ID out of range for graph")

        self.neighbors[node_id1].remove(self.node_uids[node_id2])
        self.neighbors[node_id2].remove(self.node_uids[node_id1])

    def find_index(self, value, starting_node=0):
        """Find the index of the first node that matches value in a BFS

        Performs a breadth-first search of the graph starting at node index
        starting_node.  Returns the index or None if no match is found

        Args:
            value(Node Value) - Value to match against in the graph
            starting_node(int) - Node index to start search from
        """
        if starting_node > self.node_count():
            raise IndexError("Node ID out of range.")

        node_queue = queue.LifoQueue()
        node_queue.put(starting_node)
        visited = [starting_node]
        while not node_queue.empty():
            next_id = node_queue.get()
            if (self.nodes[next_id].value == value):
                return next_id  # Success
            for uid in self.neighbors[next_id]:
                if (self.uid_to_index[uid] not in visited):
                    node_queue.put(self.uid_to_index[uid])
                    visited += [self.uid_to_index[uid]]
        return None

    def is_adjacent(self, node_id1, node_id2):
        """Test for adjacency between node1 and node2

         Args:
            node_id1(int): Index of first node
            node_id2(int): Index of second node
        """
        if node_id1 >= self.node_count() or node_id2 >= self.node_count():
            raise IndexError("Node ID out of range for graph")

        return (self.node_uids[node_id1] in self.neighbors[node_id2])

    def get_neighbors(self, node_id):
        """Return list of neighbors of the specified node

        Args:
            node_id: Index of node to examine the neighbors of

        Returns(list): List of current node IDs that are neighbors of node_id.
        """
        if node_id >= self.node_count():
            raise IndexError("Node ID out of range for graph")
        return [self.uid_to_index[i] for i in self.neighbors[node_id]]

    def shortest_path(self, node_id1, node_id2):
        """Find the shortest path between node1 and node2 on the graph

        Args:
            node_id1(int): Index of first node
            node_id2(int): Index of second node

        Returns(list): List of nodes from node_id1 to node_id2 that constitute
        the shortest possible path in the graph between those two nodes.
        """
        if node_id1 >= self.node_count() or node_id2 >= self.node_count():
            raise IndexError("Node ID out of range for graph")

        # Treat special case of equal inputs
        if node_id1 == node_id2:
            return [node_id1]

        # Initialize two arrays for backtracking the shortest path found
        previous = [None] * self.node_count()  # Tracks the moves
        distances = [None] * self.node_count()  # Records distance to nodes
        distances[node_id1] = 0

        node_queue = queue.LifoQueue()
        node_queue.put(node_id1)

        while not node_queue.empty():
            next_id = node_queue.get()
            if next_id == node_id2:
                break  # Success
            new_distance = distances[next_id] + 1
            # On each iteration, check if going to a neighbor node was
            # shortest path to that node, if so, add that node to the queue
            # and record distance and path that is being taken.
            for uid in self.neighbors[next_id]:
                if ((distances[self.uid_to_index[uid]] is None) or
                        (distances[self.uid_to_index[uid]] > new_distance)):
                    distances[self.uid_to_index[uid]] = new_distance
                    previous[self.uid_to_index[uid]] = next_id
                    node_queue.put(self.uid_to_index[uid])

        # Backtrack to get path
        if previous[node_id2] is None:
            raise IndexError("Unable to find target node, "
                             "possibly disconnected graph.")

        path = []
        next = node_id2
        while next is not None:
            path.append(next)
            next = previous[next]
        # Return reversed backtrack path to get in order from node 1 to 2
        return path[::-1]



"""Tests for ffft.py."""
import numpy
import numpy.linalg
import os
import random
import unittest

from openfermionprojectq import TimeEvolution
from openfermionprojectq._ffft import (
    ffft, fswap, fswap_adjacent, fswap_generator, apply_phase,
    fourier_transform_0, swap_adjacent_fermionic_modes, ffft_2d,
    operator_2d_fft_with_reordering)

from openfermion.hamiltonians import jellium_model
from openfermion.ops import (FermionOperator, normal_ordered,
                             QubitOperator)
from openfermion.transforms import jordan_wigner
from openfermion.utils import (count_qubits, eigenspectrum,
                               fourier_transform, Grid)

from projectq import MainEngine
from projectq.ops import (H, X, Y, Z, C, Rx, Ry, Rz, Swap, Ph,
                          Measure, All)


os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


def prepare_logical_state(register, n):
    for i in range(int(numpy.log2(n + 1)) + 1):
        if (n >> i) & 1:
            X | register[i]


def prepare_integer_fermion_operator(n):
    res = FermionOperator('')
    for i in range(int(numpy.log2(n + 1)) + 1):
        if (n >> i) & 1:
            res = FermionOperator(((i, 1),)) * res
    return res


def ordered_wavefunction(engine, indices_to_evaluate=None):
    """Return the correctly ordered wave function amplitudes.

    Args:
        engine (projectq.MainEngine): The engine from which to take the
                                      amplitudes.
        indices_to_evaluate (numpy.array): Array of integer indices to
                                           evaluate amplitudes of. If
                                           indices_to_evaluate is not
                                           set, all amplitudes are
                                           evaluated.
    """
    n_qubits = engine._qubit_idx
    if indices_to_evaluate is None:
        indices_to_evaluate = numpy.arange(2 ** n_qubits)
    elif isinstance(indices_to_evaluate, list):
        indices_to_evaluate = numpy.array(indices_to_evaluate)
    indices = numpy.zeros(len(indices_to_evaluate), dtype=int)

    # Get the qubit order dictionary and raw wavefunction from the engine.
    qubit_order_dict, unordered_wavefunction = engine.backend.cheat()

    for bit_number in range(n_qubits):
        # To each logical state for which qubit k is 1, add
        # 2 ** (qubit k's position in the raw wavefunction). This is quickly
        # given by 1 << qubit_order_dict[k].
        indices[(indices_to_evaluate & (1 << bit_number)) != 0] += (
            1 << qubit_order_dict[bit_number])

    # Return the subset of unordered_wavefunction corresponding to the indices.
    return list(map(unordered_wavefunction.__getitem__, indices))


class OrderedWavefunctionTest(unittest.TestCase):

    def setUp(self):
        self.eng1 = MainEngine()
        self.reg1 = self.eng1.allocate_qubit()
        self.eng3 = MainEngine()
        self.reg3 = self.eng3.allocate_qureg(3)

    def tearDown(self):
        All(Measure) | self.reg1
        All(Measure) | self.reg3

    def test_correct_phase_after_reordering_1qubit(self):
        H | self.reg1
        Rz(0.03) | self.reg1
        self.eng1.flush()
        angle0 = numpy.angle(ordered_wavefunction(self.eng1)[0])
        angle1 = numpy.angle(ordered_wavefunction(self.eng1)[1])
        self.assertAlmostEqual(angle1 - angle0, 0.03)

    def test_correct_phase_after_reordering_multiple_qubits(self):
        All(H) | self.reg3
        Rz(0.07) | self.reg3[1]
        self.eng3.flush()
        wavefunction = ordered_wavefunction(self.eng3)
        self.assertAlmostEqual(numpy.angle(wavefunction[1]),
                               numpy.angle(wavefunction[0]))
        self.assertAlmostEqual(numpy.angle(wavefunction[6]),
                               numpy.angle(wavefunction[2]))
        self.assertAlmostEqual(numpy.angle(wavefunction[2]) -
                               numpy.angle(wavefunction[0]),
                               0.07)

    def test_correct_resorting(self):
        """For this circuit:
        000 -> 0
        1 = 100 CZ[0, 2]-> 100 Z[1]-> 100 Swap[1, 2]-> 100 Z[0]-> -100
            Swap[0, 1]-> 010 = -2
        2 = 010 Z[1]-> -010 Swap[1, 2]-> -001 = -4
        3 = 110 Z[1]-> -110 Swap[1, 2]-> -101 -Z[0]-> 101 Swap[0, 1]-> 6
        4 = 001 Swap[1, 2]-> 010 Swap[0, 1]-> 100 = 1
        5 = 101 CZ[0, 2]-> -101 Swap[1, 2]-> -110 -Z[0]-> 110 = 3
        6 = 011 Z[1]-> -011 CZ[1, 2]-> 011 Swap[0, 1] -> 101 = 5
        7 = 111 CZ[1, 2]-> -111 Z[1]-> 111 CZ[1, 2]-> -111 Z[0]-> 111 = 7

        so the correct signs at the output are +, +, -, +, -, +, +, +.
        """
        All(H) | self.reg3
        C(Z) | (self.reg3[0], self.reg3[2])
        Z | self.reg3[1]
        Swap | (self.reg3[1], self.reg3[2])
        C(Z) | (self.reg3[1], self.reg3[2])
        Z | self.reg3[0]
        Swap | (self.reg3[1], self.reg3[0])
        self.eng3.flush()

        expected = (numpy.array([1, 1, -1, 1, -1, 1, 1, 1]) /
                    (2 * numpy.sqrt(2)))

        self.assertTrue(numpy.allclose(
            numpy.array(ordered_wavefunction(self.eng3)),
            expected))

    def test_correct_resorting_selective(self):
        """Same circuit as non-selective test:
        000 -> 0
        1 = 100 CZ[0, 2]-> 100 Z[1]-> 100 Swap[1, 2]-> 100 Z[0]-> -100
            Swap[0, 1]-> 010 = -2
        2 = 010 Z[1]-> -010 Swap[1, 2]-> -001 = -4
        3 = 110 Z[1]-> -110 Swap[1, 2]-> -101 -Z[0]-> 101 Swap[0, 1]-> 6
        4 = 001 Swap[1, 2]-> 010 Swap[0, 1]-> 100 = 1
        5 = 101 CZ[0, 2]-> -101 Swap[1, 2]-> -110 -Z[0]-> 110 = 3
        6 = 011 Z[1]-> -011 CZ[1, 2]-> 011 Swap[0, 1] -> 101 = 5
        7 = 111 CZ[1, 2]-> -111 Z[1]-> 111 CZ[1, 2]-> -111 Z[0]-> 111 = 7

        so the correct signs at the output are +, +, -, +, -, +, +, +.
        """
        All(H) | self.reg3
        C(Z) | (self.reg3[0], self.reg3[2])
        Z | self.reg3[1]
        Swap | (self.reg3[1], self.reg3[2])
        C(Z) | (self.reg3[1], self.reg3[2])
        Z | self.reg3[0]
        Swap | (self.reg3[1], self.reg3[0])
        self.eng3.flush()

        expected = numpy.array([-1, -1]) / (2 * numpy.sqrt(2))

        self.assertTrue(numpy.allclose(
            numpy.array(ordered_wavefunction(self.eng3, [2, 4])),
            expected))


class FSwapTest(unittest.TestCase):

    def setUp(self):
        self.eng = MainEngine()
        self.reg = self.eng.allocate_qureg(3)

    def tearDown(self):
        All(Measure) | self.reg

    def test_fswap_identity(self):
        fswap(self.reg, 0, 1)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([1.0] + [0.0] * 7)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_100(self):
        X | self.reg[0]
        fswap(self.reg, 0, 1)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0] * 2 + [1.0] + [0.0] * 5)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_010(self):
        X | self.reg[1]
        fswap(self.reg, 0, 1)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0, 1.0] + [0.0] * 6)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_001(self):
        X | self.reg[2]
        fswap(self.reg, 0, 1)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0] * 4 + [1.0] + [0.0] * 3)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_superposition(self):
        All(H) | self.reg
        fswap(self.reg, 0, 1)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([2. ** -1.5] * 8)
        expected[3] *= -1  # bitstring 110 gets sign-flipped by fermionic swap
        expected[7] *= -1  # bitstring 111 gets sign-flipped by fermionic swap
        self.assertTrue(numpy.allclose(ordered_wvfn, expected * -1j))

    def test_fswap_same_modes(self):
        All(H) | self.reg
        fswap(self.reg, 1, 1)  # expect identity to be applied because of this
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([2. ** -1.5] * 8)
        self.assertTrue(numpy.allclose(ordered_wvfn, expected * -1j))


class FSwapAdjacentTest(unittest.TestCase):

    def setUp(self):
        self.eng = MainEngine()
        self.reg = self.eng.allocate_qureg(3)

    def tearDown(self):
        All(Measure) | self.reg

    def test_fswap_adjacent_identity(self):
        fswap_adjacent(self.reg, 0)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([1.0] + [0.0] * 7)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_adjacent_100(self):
        X | self.reg[0]
        fswap_adjacent(self.reg, 0)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0] * 2 + [1.0] + [0.0] * 5)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_adjacent_010(self):
        X | self.reg[1]
        fswap_adjacent(self.reg, 0)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0, 1.0] + [0.0] * 6)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_adjacent_001(self):
        X | self.reg[2]
        fswap_adjacent(self.reg, 0)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([0.0] * 4 + [1.0] + [0.0] * 3)
        self.assertTrue(numpy.allclose(numpy.abs(ordered_wvfn),
                                       numpy.abs(expected)))

    def test_fswap_adjacent_superposition(self):
        All(H) | self.reg
        fswap_adjacent(self.reg, 0)
        self.eng.flush()

        ordered_wvfn = numpy.array(ordered_wavefunction(self.eng))
        expected = numpy.array([2. ** -1.5] * 8)
        expected[3] *= -1  # bitstring 110 gets sign-flipped by fermionic swap
        expected[7] *= -1  # bitstring 111 gets sign-flipped by fermionic swap
        self.assertTrue(numpy.allclose(ordered_wvfn, expected * -1j))


class ApplyPhaseTest(unittest.TestCase):

    def setUp(self):
        self.eng = MainEngine()
        self.reg = self.eng.allocate_qureg(3)

    def tearDown(self):
        All(Measure) | self.reg

    def test_apply_phase_1qubit(self):
        eng = MainEngine()
        reg = eng.allocate_qubit()
        H | reg
        apply_phase(reg, 0, 0.03)
        eng.flush()
        self.assertAlmostEqual(numpy.angle(ordered_wavefunction(eng)[1]) -
                               numpy.angle(ordered_wavefunction(eng)[0]),
                               0.06)
        Measure | reg

    def test_apply_phase_multiple_qubits(self):
        All(H) | self.reg
        apply_phase(self.reg, 1, 0.07)
        self.eng.flush()
        wavefunction = ordered_wavefunction(self.eng)
        self.assertAlmostEqual(numpy.angle(wavefunction[1]),
                               numpy.angle(wavefunction[0]))
        self.assertAlmostEqual(numpy.angle(wavefunction[6]),
                               numpy.angle(wavefunction[2]))
        self.assertAlmostEqual(numpy.angle(wavefunction[2]) -
                               numpy.angle(wavefunction[0]),
                               0.14)


class FFFTNModeIntegrationTest(unittest.TestCase):

    def test_ffft_0mode_error(self):
        eng = MainEngine()
        reg = eng.allocate_qureg(1)
        with self.assertRaises(ValueError):
            ffft(eng, reg, 0)

    def test_ffft_1mode_error(self):
        eng = MainEngine()
        reg = eng.allocate_qureg(1)
        with self.assertRaises(ValueError):
            ffft(eng, reg, 1)

    def test_ffft_2modes_properly_applied(self):
        eng_ft_0 = MainEngine()
        reg_ft_0 = eng_ft_0.allocate_qureg(2)

        eng_ffft_recursive = MainEngine()
        reg_ffft_recursive = eng_ffft_recursive.allocate_qureg(2)

        ffft(eng_ffft_recursive, reg_ffft_recursive, 2)
        fourier_transform_0(reg_ft_0, 0, 1)
        eng_ffft_recursive.flush()
        eng_ft_0.flush()

        All(Measure) | reg_ffft_recursive
        All(Measure) | reg_ft_0

        self.assertTrue(numpy.allclose(
            ordered_wavefunction(eng_ffft_recursive),
            ordered_wavefunction(eng_ft_0)))

    def test_2mode_ffft_correct_frequencies(self):
        n_qubits = 2
        frequencies_seen = numpy.zeros(n_qubits)

        for qubit in range(n_qubits):
            engine = MainEngine()
            register = engine.allocate_qureg(n_qubits)
            X | register[qubit]
            ffft(engine, register, n_qubits)
            engine.flush()

            wavefunction = ordered_wavefunction(engine)
            nonzero_wavefunction_elmts = []
            for el in wavefunction:
                if abs(el) > 10 ** -5:
                    nonzero_wavefunction_elmts.append(el)
            All(Measure) | register

            phase_factor = (nonzero_wavefunction_elmts[1] /
                            nonzero_wavefunction_elmts[0])
            offset = numpy.angle(nonzero_wavefunction_elmts[0])

            self.assertAlmostEqual(offset, 0.0)
            for i in range(1, len(nonzero_wavefunction_elmts)):
                self.assertAlmostEqual(phase_factor,
                                       (nonzero_wavefunction_elmts[i] /
                                        nonzero_wavefunction_elmts[i - 1]))
            frequencies_seen[qubit] = numpy.angle(phase_factor)

        frequencies_seen = numpy.sort(frequencies_seen)
        expected = numpy.sort(2 * numpy.pi * numpy.arange(n_qubits) / n_qubits)

        self.assertTrue(numpy.allclose(frequencies_seen, expected))

    def test_4mode_ffft_correct_frequencies(self):
        n_qubits = 4
        frequencies_seen = numpy.zeros(n_qubits)
        offset = None

        for qubit in range(n_qubits):
            engine = MainEngine()
            register = engine.allocate_qureg(n_qubits)
            X | register[qubit]
            ffft(engine, register, n_qubits)
            engine.flush()

            wavefunction = ordered_wavefunction(engine)
            nonzero_wavefunction_elmts = []
            for el in wavefunction:
                if abs(el) > 10 ** -5:
                    nonzero_wavefunction_elmts.append(el)
            All(Measure) | register
            engine.flush()

            phase_factor = (nonzero_wavefunction_elmts[1] /
                            nonzero_wavefunction_elmts[0])
            if offset is None:
                offset = numpy.angle(nonzero_wavefunction_elmts[0])
            self.assertAlmostEqual(offset,
                                   numpy.angle(nonzero_wavefunction_elmts[0]))

            for i in range(1, len(nonzero_wavefunction_elmts)):
                self.assertAlmostEqual(phase_factor,
                                       (nonzero_wavefunction_elmts[i] /
                                        nonzero_wavefunction_elmts[i - 1]))
            frequencies_seen[qubit] = numpy.angle(phase_factor)
            if frequencies_seen[qubit] < -(10 ** -5):
                frequencies_seen[qubit] += 2 * numpy.pi

        frequencies_seen = numpy.sort(frequencies_seen)
        expected = numpy.sort(2 * numpy.pi * numpy.arange(n_qubits) / n_qubits)

        self.assertTrue(numpy.allclose(frequencies_seen, expected))

    def test_8mode_ffft_correct_frequencies(self):
        n_qubits = 8
        frequencies_seen = numpy.zeros(n_qubits)
        offset = None

        for qubit in range(n_qubits):
            engine = MainEngine()
            register = engine.allocate_qureg(n_qubits)
            X | register[qubit]
            ffft(engine, register, n_qubits)
            engine.flush()

            wavefunction = ordered_wavefunction(engine)
            nonzero_wavefunction_elmts = []
            for el in wavefunction:
                if abs(el) > 10 ** -5:
                    nonzero_wavefunction_elmts.append(el)

            All(Measure) | register

            phase_factor = (nonzero_wavefunction_elmts[1] /
                            nonzero_wavefunction_elmts[0])
            if offset is None:
                offset = numpy.angle(nonzero_wavefunction_elmts[0])
            self.assertAlmostEqual(offset,
                                   numpy.angle(nonzero_wavefunction_elmts[0]))

            for i in range(1, len(nonzero_wavefunction_elmts)):
                self.assertAlmostEqual(phase_factor,
                                       (nonzero_wavefunction_elmts[i] /
                                        nonzero_wavefunction_elmts[i - 1]),
                                       msg="wvfn = {}".format(
                                           nonzero_wavefunction_elmts))
            frequencies_seen[qubit] = numpy.angle(phase_factor)
            if frequencies_seen[qubit] < -(10 ** -5):
                frequencies_seen[qubit] += 2 * numpy.pi

        frequencies_seen = numpy.sort(frequencies_seen)
        expected = numpy.sort(2 * numpy.pi * numpy.arange(n_qubits) / n_qubits)

        self.assertTrue(numpy.allclose(frequencies_seen, expected))


class SwapAdjacentFermionicModesTest(unittest.TestCase):

    def test_basic_swap(self):
        operator = normal_ordered(FermionOperator('2^ 2 4^ 4'))
        operator_swapped = swap_adjacent_fermionic_modes(operator, 2)

        self.assertTrue(operator_swapped == FermionOperator('4^ 3^ 4 3', -1.0))


class FFFTPlaneWaveIntegrationTest(unittest.TestCase):

    def setUp(self):
        random.seed(17)

    def test_4mode_ffft_with_external_swaps_all_logical_states(self):
        n_qubits = 4

        grid = Grid(dimensions=1, length=n_qubits, scale=1.0)

        for i in range(2 ** n_qubits):
            eng = MainEngine()
            register = eng.allocate_qureg(n_qubits)
            prepare_logical_state(register, i)

            ffft(eng, register, n_qubits)
            Ph(3 * numpy.pi / 4) | register[0]
            eng.flush()
            wvfn = ordered_wavefunction(eng)

            fermion_operator = prepare_integer_fermion_operator(i)

            # Reorder the modes for correct input to the FFFT.
            # Swap 0123 to 2301 for fourier_transform. Additionally, the
            # FFFT's ordering is 0213, so connect 0213 -> 0123 -> 2301.
            swap_mode_list = [1] + [1, 0, 2, 1]
            for mode in swap_mode_list:
                fermion_operator = normal_ordered(fermion_operator)
                fermion_operator = swap_adjacent_fermionic_modes(
                    fermion_operator, mode)

            ffft_result = fourier_transform(fermion_operator,
                                            grid, spinless=True)
            ffft_result = normal_ordered(ffft_result)

            swap_mode_list = [1, 0, 2, 1]  # After FFFT, swap 2301 -> 0123
            for mode in swap_mode_list:
                ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)

            converted_wvfn = numpy.zeros(2 ** n_qubits, dtype=complex)
            for term in ffft_result.terms:
                index = sum(2 ** site[0] for site in term)
                converted_wvfn[index] = ffft_result.terms[term]

            All(Measure) | register

            self.assertTrue(numpy.allclose(wvfn, converted_wvfn))

    def test_8mode_ffft_with_external_swaps_on_single_logical_state(self):
        n_qubits = 8
        grid = Grid(dimensions=1, length=n_qubits, scale=1.0)

        eng = MainEngine()
        register = eng.allocate_qureg(n_qubits)

        state_index = 157

        prepare_logical_state(register, state_index)

        ffft(eng, register, n_qubits)
        Ph(3 * numpy.pi / 4) | register[0]
        eng.flush()
        wvfn = ordered_wavefunction(eng)

        fermion_operator = prepare_integer_fermion_operator(state_index)

        # Swap 01234567 to 45670123 for fourier_transform. Additionally,
        # the FFFT's ordering is 04261537, so swap 04261537 to 01234567,
        # and then 01234567 to 45670123.
        swap_mode_list = ([1, 3, 5, 2, 4, 1, 3, 5] +
                          [3, 2, 4, 1, 3, 5, 0, 2, 4, 6, 1, 3, 5, 2, 4, 3])
        for mode in swap_mode_list:
            fermion_operator = normal_ordered(fermion_operator)
            fermion_operator = swap_adjacent_fermionic_modes(
                fermion_operator, mode)

        ffft_result = fourier_transform(fermion_operator,
                                        grid, spinless=True)
        ffft_result = normal_ordered(ffft_result)

        # After the FFFT, swap 45670123 -> 01234567.
        swap_mode_list = [3, 2, 4, 1, 3, 5, 0, 2, 4, 6, 1, 3, 5, 2, 4, 3]
        for mode in swap_mode_list:
            ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)

        converted_wvfn = numpy.zeros(2 ** n_qubits, dtype=complex)
        for term in ffft_result.terms:
            index = sum(2 ** site[0] for site in term)
            converted_wvfn[index] = ffft_result.terms[term]

        All(Measure) | register

        self.assertTrue(numpy.allclose(wvfn, converted_wvfn))

    def test_4mode_ffft_with_external_swaps_equal_expectation_values(self):
        n_qubits = 4

        grid = Grid(dimensions=1, length=n_qubits, scale=1.0)
        dual_basis = jellium_model(grid, spinless=True, plane_wave=False)
        ffft_result = normal_ordered(dual_basis)

        # Reorder the modes for correct input to the FFFT.
        # Swap 0123 to 2301 for fourier_transform. Additionally, the
        # FFFT's ordering is 0213, so connect 0213 -> 0123 -> 2301.
        swap_mode_list = [1] + [1, 0, 2, 1]
        for mode in swap_mode_list:
            ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)
            ffft_result = normal_ordered(ffft_result)

        ffft_result = fourier_transform(ffft_result,
                                        grid, spinless=True)
        ffft_result = normal_ordered(ffft_result)

        swap_mode_list = [1, 0, 2, 1]  # After FFFT, swap 2301 -> 0123
        for mode in swap_mode_list:
            ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)
            ffft_result = normal_ordered(ffft_result)

        jw_dual_basis = jordan_wigner(dual_basis)
        jw_plane_wave = jordan_wigner(ffft_result)

        # Do plane wave and dual basis calculations simultaneously.
        pw_engine = MainEngine()
        pw_wavefunction = pw_engine.allocate_qureg(n_qubits)
        pw_engine.flush()
        db_engine = MainEngine()
        db_wavefunction = db_engine.allocate_qureg(n_qubits)
        db_engine.flush()

        # Choose random state.
        state = numpy.zeros(2 ** n_qubits, dtype=complex)
        for i in range(len(state)):
            state[i] = (random.random() *
                        numpy.exp(1j * 2 * numpy.pi * random.random()))
        state /= numpy.linalg.norm(state)

        # Put randomly chosen state in the registers.
        pw_engine.backend.set_wavefunction(state, pw_wavefunction)
        db_engine.backend.set_wavefunction(state, db_wavefunction)

        prepare_logical_state(pw_wavefunction, i)
        prepare_logical_state(db_wavefunction, i)

        All(H) | [pw_wavefunction[1], pw_wavefunction[3]]
        All(H) | [db_wavefunction[1], db_wavefunction[3]]

        ffft(db_engine, db_wavefunction, n_qubits)
        Ph(3 * numpy.pi / 4) | db_wavefunction[0]

        # Flush the engine and compute expectation values and eigenvalues.
        pw_engine.flush()
        db_engine.flush()

        plane_wave_expectation_value = (
            pw_engine.backend.get_expectation_value(
                jw_dual_basis, pw_wavefunction))
        dual_basis_expectation_value = (
            db_engine.backend.get_expectation_value(
                jw_plane_wave, db_wavefunction))

        All(Measure) | pw_wavefunction
        All(Measure) | db_wavefunction

        self.assertAlmostEqual(plane_wave_expectation_value,
                               dual_basis_expectation_value)

    def test_8mode_ffft_with_external_swaps_equal_expectation_values(self):
        n_qubits = 8

        grid = Grid(dimensions=1, length=n_qubits, scale=1.0)
        dual_basis = jellium_model(grid, spinless=True, plane_wave=False)
        ffft_result = normal_ordered(dual_basis)

        # Swap 01234567 to 45670123 for fourier_transform. Additionally,
        # the FFFT's ordering is 04261537, so swap 04261537 to 01234567,
        # and then 01234567 to 45670123.
        swap_mode_list = ([1, 3, 5, 2, 4, 1, 3, 5] +
                          [3, 2, 4, 1, 3, 5, 0, 2, 4, 6, 1, 3, 5, 2, 4, 3])
        for mode in swap_mode_list:
            ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)
            ffft_result = normal_ordered(ffft_result)

        ffft_result = fourier_transform(ffft_result,
                                        grid, spinless=True)
        ffft_result = normal_ordered(ffft_result)

        # After the FFFT, swap 45670123 -> 01234567.
        swap_mode_list = [3, 2, 4, 1, 3, 5, 0, 2, 4, 6, 1, 3, 5, 2, 4, 3]
        for mode in swap_mode_list:
            ffft_result = swap_adjacent_fermionic_modes(ffft_result, mode)
            ffft_result = normal_ordered(ffft_result)

        jw_dual_basis = jordan_wigner(dual_basis)
        jw_plane_wave = jordan_wigner(ffft_result)

        # Do plane wave and dual basis calculations simultaneously.
        pw_engine = MainEngine()
        pw_wavefunction = pw_engine.allocate_qureg(n_qubits)
        db_engine = MainEngine()
        db_wavefunction = db_engine.allocate_qureg(n_qubits)

        pw_engine.flush()
        db_engine.flush()

        # Choose random state.
        state = numpy.zeros(2 ** n_qubits, dtype=complex)
        for i in range(len(state)):
            state[i] = (random.random() *
                        numpy.exp(1j * 2 * numpy.pi * random.random()))
        state /= numpy.linalg.norm(state)

        # Put randomly chosen state in the registers.
        pw_engine.backend.set_wavefunction(state, pw_wavefunction)
        db_engine.backend.set_wavefunction(state, db_wavefunction)

        prepare_logical_state(pw_wavefunction, i)
        prepare_logical_state(db_wavefunction, i)

        All(H) | [pw_wavefunction[1], pw_wavefunction[3]]
        All(H) | [db_wavefunction[1], db_wavefunction[3]]

        ffft(db_engine, db_wavefunction, n_qubits)
        Ph(3 * numpy.pi / 4) | db_wavefunction[0]

        # Flush the engine and compute expectation values and eigenvalues.
        pw_engine.flush()
        db_engine.flush()

        plane_wave_expectation_value = (
            pw_engine.backend.get_expectation_value(
                jw_dual_basis, pw_wavefunction))
        dual_basis_expectation_value = (
            db_engine.backend.get_expectation_value(
                jw_plane_wave, db_wavefunction))

        All(Measure) | pw_wavefunction
        All(Measure) | db_wavefunction

        self.assertAlmostEqual(plane_wave_expectation_value,
                               dual_basis_expectation_value)

    def test_ffft_2d_4x4_logical_state(self):
        system_size = 4
        grid = Grid(dimensions=2, length=system_size, scale=1.0)

        fermion_operator = FermionOperator('12^ 11^ 4^ 3^')
        ffft_result = operator_2d_fft_with_reordering(
            fermion_operator, system_size)

        eng = MainEngine()
        reg = eng.allocate_qureg(system_size ** 2)
        X | reg[3]
        X | reg[4]
        X | reg[11]
        X | reg[12]
        ffft_2d(eng, reg, system_size)
        eng.flush()

        engine_wvfn = ordered_wavefunction(eng)
        operator_wvfn = numpy.zeros(2 ** (system_size ** 2), dtype=complex)
        for term in ffft_result.terms:
            i = sum(2 ** site[0] for site in term)
            operator_wvfn[i] = ffft_result.terms[term]

        All(Measure) | reg

        self.assertTrue(numpy.allclose(engine_wvfn, operator_wvfn))

    def test_ffft_2d_4x4_equal_expectation_values(self):
        system_size = 4
        n_qubits = 16
        grid = Grid(dimensions=2, length=system_size, scale=1.0)
        dual_basis = jellium_model(grid, spinless=True, plane_wave=False)
        ffft_result = operator_2d_fft_with_reordering(dual_basis, system_size)

        jw_dual_basis = jordan_wigner(dual_basis)
        jw_plane_wave = jordan_wigner(ffft_result)

        # Do plane wave and dual basis calculations together.
        pw_engine = MainEngine()
        pw_wavefunction = pw_engine.allocate_qureg(system_size ** 2)
        db_engine = MainEngine()
        db_wavefunction = db_engine.allocate_qureg(system_size ** 2)

        pw_engine.flush()
        db_engine.flush()

        # Choose random state.
        state = numpy.zeros(2 ** n_qubits, dtype=complex)
        for i in range(len(state)):
            state[i] = (random.random() *
                        numpy.exp(1j * 2 * numpy.pi * random.random()))
        state /= numpy.linalg.norm(state)

        # Put randomly chosen state in the registers.
        pw_engine.backend.set_wavefunction(state, pw_wavefunction)
        db_engine.backend.set_wavefunction(state, db_wavefunction)

        # Apply the FFFT to the dual basis wave function.
        ffft_2d(db_engine, db_wavefunction, system_size)

        # Flush the engine and compute expectation values and eigenvalues.
        pw_engine.flush()
        db_engine.flush()

        plane_wave_expectation_value = (
            pw_engine.backend.get_expectation_value(
                jw_dual_basis, pw_wavefunction))
        dual_basis_expectation_value = (
            db_engine.backend.get_expectation_value(
                jw_plane_wave, db_wavefunction))

        All(Measure) | pw_wavefunction
        All(Measure) | db_wavefunction

        self.assertAlmostEqual(plane_wave_expectation_value,
                               dual_basis_expectation_value)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for projectq.ops._time_evolution."""
import cmath
import copy

import numpy
import pytest

from openfermion.ops import QubitOperator
from openfermionprojectq import _time_evolution as te

from projectq import MainEngine
from projectq.cengines import DummyEngine
from projectq.ops import BasicGate, NotMergeable, Ph


@pytest.mark.parametrize("coefficient", [0.5, numpy.float64(2.303)])
def test_time_evolution_init_int_time(coefficient):
    hamiltonian = coefficient * QubitOperator("X0 Z1")
    hamiltonian += QubitOperator("Z2", 0.5)
    gate1 = te.TimeEvolution(2, hamiltonian)
    assert gate1.hamiltonian.isclose(hamiltonian)
    assert gate1.time == 2


@pytest.mark.parametrize("coefficient", [0.5, numpy.float64(2.303)])
def test_init_float_time(coefficient):
    hamiltonian = coefficient * QubitOperator("X0 Z1")
    hamiltonian += QubitOperator("Z2", 0.5)
    gate2 = te.TimeEvolution(2.1, hamiltonian)
    assert gate2.hamiltonian.isclose(hamiltonian)
    assert gate2.time == pytest.approx(2.1)


def test_init_makes_copy():
    hamiltonian = QubitOperator("X0 Z1")
    gate = te.TimeEvolution(2.1, hamiltonian)
    hamiltonian = None
    assert gate.hamiltonian is not None


def test_init_cast_complex_to_float():
    hamiltonian = QubitOperator("Z2", 2+0j)
    gate = te.TimeEvolution(1, hamiltonian)
    assert isinstance(gate.hamiltonian.terms[((2, 'Z'),)], float)
    pytest.approx(gate.hamiltonian.terms[((2, 'Z'),)]) == 2.0


def test_init_negative_time():
    hamiltonian = QubitOperator("Z2", 2)
    gate = te.TimeEvolution(-1, hamiltonian)
    assert gate.time == -1


def test_get_inverse():
    hamiltonian = QubitOperator("Z2", 2)
    gate = te.TimeEvolution(2, hamiltonian)
    inverse = gate.get_inverse()
    assert gate.time == 2
    assert gate.hamiltonian.isclose(hamiltonian)
    assert inverse.time == -2
    assert inverse.hamiltonian.isclose(hamiltonian)


def test_get_merged_one_term():
    hamiltonian = QubitOperator("Z2", 2)
    gate = te.TimeEvolution(2, hamiltonian)
    hamiltonian2 = QubitOperator("Z2", 4)
    gate2 = te.TimeEvolution(5, hamiltonian2)
    merged = gate.get_merged(gate2)
    # This is not a requirement, the hamiltonian could also be the other
    # if we change implementation
    assert merged.hamiltonian.isclose(hamiltonian)
    assert merged.time == pytest.approx(12)


def test_get_merged_multiple_terms():
    hamiltonian = QubitOperator("Z2", 2)
    hamiltonian += QubitOperator("X3", 1)
    gate = te.TimeEvolution(2, hamiltonian)
    hamiltonian2 = QubitOperator("Z2", 4)
    hamiltonian2 += QubitOperator("X3", 2 + 1e-10)
    gate2 = te.TimeEvolution(5, hamiltonian2)
    merged = gate.get_merged(gate2)
    # This is not a requirement, the hamiltonian could also be the other
    # if we change implementation
    assert merged.hamiltonian.isclose(hamiltonian)
    assert merged.time == pytest.approx(12)


def test_get_merged_not_close_enough():
    hamiltonian = QubitOperator("Z2", 2)
    hamiltonian += QubitOperator("X3", 1)
    gate = te.TimeEvolution(2, hamiltonian)
    hamiltonian2 = QubitOperator("Z2", 4)
    hamiltonian2 += QubitOperator("X3", 2+1e-8)
    gate2 = te.TimeEvolution(5, hamiltonian2)
    with pytest.raises(NotMergeable):
        merged = gate.get_merged(gate2)


def test_get_merged_bad_gate():
    hamiltonian = QubitOperator("Z2", 2)
    gate = te.TimeEvolution(2, hamiltonian)
    other = BasicGate()
    with pytest.raises(NotMergeable):
        gate.get_merged(other)


def test_get_merged_different_hamiltonian():
    hamiltonian = QubitOperator("Z2", 2)
    gate = te.TimeEvolution(2, hamiltonian)
    hamiltonian2 = QubitOperator("Y2", 2)
    gate2 = te.TimeEvolution(2, hamiltonian2)
    with pytest.raises(NotMergeable):
        gate.get_merged(gate2)


def test_or_one_qubit():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qubit = eng.allocate_qubit()
    hamiltonian = QubitOperator("Z0", 2)
    te.TimeEvolution(2.1, hamiltonian) | qubit[0]
    te.TimeEvolution(3, hamiltonian) | (qubit[0],)
    eng.flush()
    cmd1 = saving_backend.received_commands[1]
    assert cmd1.gate.hamiltonian.isclose(hamiltonian)
    assert cmd1.gate.time == pytest.approx(2.1)
    assert len(cmd1.qubits) == 1 and len(cmd1.qubits[0]) == 1
    assert cmd1.qubits[0][0].id == qubit[0].id
    cmd2 = saving_backend.received_commands[2]
    assert cmd2.gate.hamiltonian.isclose(hamiltonian)
    assert cmd2.gate.time == pytest.approx(3)
    assert len(cmd2.qubits) == 1 and len(cmd2.qubits[0]) == 1
    assert cmd2.qubits[0][0].id == qubit[0].id


def test_or_one_qureg():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(5)
    hamiltonian = QubitOperator("X0 Z4", 2)
    te.TimeEvolution(2.1, hamiltonian) | qureg
    te.TimeEvolution(3, hamiltonian) | (qureg,)
    eng.flush()
    rescaled_h = QubitOperator("X0 Z1", 2)
    cmd1 = saving_backend.received_commands[5]
    assert cmd1.gate.hamiltonian.isclose(rescaled_h)
    assert cmd1.gate.time == pytest.approx(2.1)
    assert len(cmd1.qubits) == 1 and len(cmd1.qubits[0]) == 2
    assert cmd1.qubits[0][0].id == qureg[0].id
    assert cmd1.qubits[0][1].id == qureg[4].id
    cmd2 = saving_backend.received_commands[6]
    assert cmd2.gate.hamiltonian.isclose(rescaled_h)
    assert cmd2.gate.time == pytest.approx(3)
    assert len(cmd2.qubits) == 1 and len(cmd2.qubits[0]) == 2
    assert cmd2.qubits[0][0].id == qureg[0].id
    assert cmd2.qubits[0][1].id == qureg[4].id


def test_or_two_qubits_error():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(2)
    hamiltonian = QubitOperator("Z0", 2)
    with pytest.raises(TypeError):
        te.TimeEvolution(2.1, hamiltonian) | (qureg[0], qureg[1])


def test_or_two_quregs_error():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(2)
    qureg2 = eng.allocate_qureg(2)
    hamiltonian = QubitOperator("Z0", 2)
    with pytest.raises(TypeError):
        te.TimeEvolution(2.1, hamiltonian) | (qureg, qureg2)


def test_or_not_enough_qubits():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(2)
    hamiltonian = QubitOperator("Z0 X3", 2)
    with pytest.raises(ValueError):
        te.TimeEvolution(2.1, hamiltonian) | qureg


def test_or_multiple_terms():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(4)
    hamiltonian = QubitOperator("X0 Z3", 2)
    hamiltonian += QubitOperator("Y1", 0.5)
    te.TimeEvolution(2.1, hamiltonian) | qureg
    eng.flush()
    rescaled_h = QubitOperator("X0 Z2", 2)
    rescaled_h += QubitOperator("Y1", 0.5)
    cmd1 = saving_backend.received_commands[4]
    assert cmd1.gate.hamiltonian.isclose(rescaled_h)
    assert cmd1.gate.time == pytest.approx(2.1)
    assert len(cmd1.qubits) == 1 and len(cmd1.qubits[0]) == 3
    assert cmd1.qubits[0][0].id == qureg[0].id
    assert cmd1.qubits[0][1].id == qureg[1].id
    assert cmd1.qubits[0][2].id == qureg[3].id


def test_or_gate_not_mutated():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(4)
    hamiltonian = QubitOperator("X0 Z3", 2)
    hamiltonian += QubitOperator("Y1", 0.5)
    correct_h = copy.deepcopy(hamiltonian)
    gate = te.TimeEvolution(2.1, hamiltonian)
    gate | qureg
    eng.flush()
    assert gate.hamiltonian.isclose(correct_h)
    assert gate.time == pytest.approx(2.1)


def test_or_gate_identity():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    qureg = eng.allocate_qureg(4)
    hamiltonian = QubitOperator((), 3.4)
    correct_h = copy.deepcopy(hamiltonian)
    gate = te.TimeEvolution(2.1, hamiltonian)
    gate | qureg
    eng.flush()
    cmd = saving_backend.received_commands[4]
    assert isinstance(cmd.gate, Ph)
    assert cmd.gate == Ph(-3.4 * 2.1)
    correct = numpy.array([[cmath.exp(-1j * 3.4 * 2.1), 0],
                           [0, cmath.exp(-1j * 3.4 * 2.1)]])
    print(correct)
    print(cmd.gate.matrix)
    assert numpy.allclose(cmd.gate.matrix, correct)


def test_eq_not_implemented():
    hamiltonian = QubitOperator("X0 Z1")
    gate = te.TimeEvolution(2.1, hamiltonian)
    assert gate.__eq__("0") == NotImplemented


def test_ne_not_implemented():
    hamiltonian = QubitOperator("X0 Z1")
    gate = te.TimeEvolution(2.1, hamiltonian)
    assert gate.__ne__("0") == NotImplemented


def test_str():
    hamiltonian = QubitOperator("X0 Z1")
    hamiltonian += QubitOperator("Y1", 0.5)
    gate = te.TimeEvolution(2.1, hamiltonian)
    assert (str(gate) == "exp(-2.1j * (0.5 Y1 +\n1.0 X0 Z1))" or
            str(gate) == "exp(-2.1j * (1.0 X0 Z1 +\n0.5 Y1))")



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests the code in the examples directory of the git repo."""
import nbformat
import os
import subprocess
import sys
import tempfile
import unittest


class ExampleTest(unittest.TestCase):

    def setUp(self):
        this_directory = os.path.realpath(__file__)
        string_length = len(this_directory)
        directory = this_directory[:(string_length - 33)] + '/examples/'
        demo_name = 'openfermionprojectq_demo.ipynb'
        self.path = directory + demo_name

    def test_demo(self):
        """Execute a notebook via nbconvert and collect output."""

        # Determine if python 2 or 3 is being used.
        major_version, minor_version = sys.version_info[:2]
        if major_version == 2 or minor_version == 6:
            version = str(major_version)

            # Run ipython notebook.
            with tempfile.NamedTemporaryFile(suffix='.ipynb') as output_file:
                args = ['jupyter',
                        'nbconvert',
                        '--to',
                        'notebook',
                        '--execute',
                        '--ExecutePreprocessor.timeout=600',
                        '--ExecutePreprocessor.kernel_name=python{}'.format(
                            version),
                        '--output',
                        output_file.name,
                        self.path]
                subprocess.check_call(args)
                output_file.seek(0)
                nb = nbformat.read(output_file, nbformat.current_nbformat)

            # Parse output and make sure there are no errors.
            errors = [output for cell in nb.cells if "outputs" in cell for
                      output in cell["outputs"] if
                      output.output_type == "error"]
        else:
            errors = []
        self.assertEqual(errors, [])



"""Functions to perform the fermionic fast Fourier transform."""
import itertools
import numpy

from openfermionprojectq import TimeEvolution
from openfermionprojectq._parallel_bubble_sort import (
    index_of_position_in_1d_array, parallel_bubble_sort)
from openfermion.ops import (FermionOperator, normal_ordered,
                             QubitOperator)
from openfermion.transforms import jordan_wigner
from openfermion.utils import fourier_transform, Grid

from projectq import MainEngine
from projectq.ops import (H, X, Y, Z, C, Rx, Ry, Rz, Ph, Swap,
                          Measure, All)


def fswap_generator(mode_a, mode_b):
    """Give the two-qubit fermionic swap operator between two modes.

    Args:
        mode_a, mode_b (int): The two modes to be swapped.

    Returns:
        FermionOperator which swaps modes mode_a and mode_b.

    Notes:
        A fermionic swap is the same as the usual swap, with the
        difference that it applies a sign -1 to the state |11>.
    """
    return (FermionOperator('') +
            FermionOperator(((mode_a, 1), (mode_b, 0))) +
            FermionOperator(((mode_b, 1), (mode_a, 0))) -
            FermionOperator(((mode_a, 1), (mode_a, 0))) -
            FermionOperator(((mode_b, 1), (mode_b, 0))))


def fswap(register, mode_a, mode_b, fermion_to_spin_mapping=jordan_wigner):
    """Apply the fermionic swap operator to two modes.

    The fermionic swap is applied to the qubits corresponding to mode_a
    and mode_b, after the Jordan-Wigner transformation.

    Args:
        register (projectq.Qureg): The register of qubits to act on.
        mode_a, mode_b (int): The two modes to swap.
        fermion_to_spin_mapping (function): Transformation from fermionic
            to spin operators to use. Defaults to jordan_wigner.
    """
    operator = fswap_generator(mode_a, mode_b)
    TimeEvolution(numpy.pi / 2., fermion_to_spin_mapping(operator)) | register


def fswap_adjacent(register, mode):
    """Apply the fermionic swap operator to adjacent modes.

    The fermionic swap is applied to the qubits corresponding to mode
    and mode + 1, after the Jordan-Wigner transformation.

    Args:
        register (projectq.Qureg): The register of qubits to act on.
        mode (int): The lower of the two modes to apply the phase to.

    Notes:
        fswap_adjacent is a slightly optimized version of fswap for a
        more restricted case. Because the modes are adjacent, the
        Jordan-Wigner transform doesn't need to be computed.
    """
    operator = (QubitOperator(((mode, 'Z'),)) +
                QubitOperator(((mode + 1, 'Z'),)) +
                QubitOperator(((mode, 'X'), (mode + 1, 'X'))) +
                QubitOperator(((mode, 'Y'), (mode + 1, 'Y'))))
    TimeEvolution(numpy.pi / 4., operator) | register


def apply_phase(register, mode, phase):
    """Apply a phase to a fermionic mode if it is occupied.

    The phase is applied to the qubit corresponding to the mode under
    the Jordan-Wigner transformation.

    Args:
        register (projectq.Qureg): The register of qubits to act on.
        mode (int): The mode to apply the phase to.
    """
    Rz(2 * phase) | register[mode]


def fourier_transform_0_generator(mode_a, mode_b):
    """Give the generator of the two-qubit fermionic Fourier transform.

    Args:
        mode_a, mode_b (int): The two modes to Fourier transform.

    Notes:
      This is equivalent to a Hadamard transform on the singly-occupied
      space. It converts |01> to (|01> + |10>) / sqrt(2) and |10> to
      (|01> - |10>) / sqrt(2). Additionally, it applies a sign -1 to
      the state |11>. The state |00> is not changed.
    """
    xy = ((FermionOperator(((mode_a, 1),)) -
           1j * FermionOperator(((mode_a, 0),))) *
          (FermionOperator(((mode_b, 1),)) -
           1j * FermionOperator(((mode_b, 0),))))

    yx = ((FermionOperator(((mode_a, 1),)) +
           1j * FermionOperator(((mode_a, 0),))) *
          (FermionOperator(((mode_b, 1),)) +
           1j * FermionOperator(((mode_b, 0),))))

    return xy - yx


def fourier_transform_0(register, mode_a, mode_b):
    """Apply the fermionic Fourier transform to two modes.

    The fermionic Fourier transform is applied to the qubits
    corresponding to mode_a and mode_b, , after the Jordan-Wigner

    Args:
        register (projectq.Qureg): The register of qubits to act on.
        mode_a, mode_b (int): The two modes to Fourier transform.
    """
    operator = fourier_transform_0_generator(mode_a, mode_b)
    jw_operator = jordan_wigner(operator)
    Z | register[mode_b]
    TimeEvolution(numpy.pi / 8., jw_operator) | register


def fourier_transform_0_adjacent(register, mode):
    """Apply the fermionic Fourier transform to two adjacent modes.

    The fermionic Fourier transform is applied to the qubits
    corresponding to mode and mode + 1 after the Jordan-Wigner
    transformation.

    Args:
        register (projectq.Qureg): The register of qubits to act on.
        mode (int): The lower of the two modes to Fourier transform.
    """
    jw_operator = (QubitOperator(((mode, 'X'), (mode + 1, 'Y'))) -
                   QubitOperator(((mode, 'Y'), (mode + 1, 'X'))))
    Z | register[mode + 1]
    TimeEvolution(numpy.pi / 8., jw_operator) | register


def ffft(engine, register, n, start=0):
    """Apply the n-mode fermionic fast Fourier transform to the given
    register on qubits start to start + n (not inclusive).

    Args:
        engine (projectq.MainEngine): The simulator engine with the
                                      register.
        register (projectq.QuReg): The register to apply the FFFT to.
        n (int): The number of modes in the system.
        start (int): The starting mode for the FFFT. The FFFT acts on
                     modes start through start + n (not inclusive).
                     Defaults to 0.

    Notes:
        This algorithm uses radix-2 decimation-in-time, so n must be a
        binary power. This decimation is head recursive (the 2-mode FFFT
        is applied as the first part of the 4-mode FFFT, which is applied
        as the first part of the 8-mode FFFT, etc).
    """
    if n < 2 or (n & (n - 1)):
        raise ValueError('n must be a binary power.')

    if n == 2:
        fourier_transform_0_adjacent(register, start)
    else:
        ffft(engine, register, n // 2, start)
        ffft(engine, register, n // 2, start + n // 2)

        # Apply initial swap network.
        center = start + n // 2
        for i in range(n // 2):
            for j in range(i):
                # fswap is equivalent to Swap followed by C(Z)
                fswap_adjacent(register, center + 2*j - i)

        # Apply phases and two-qubit Fourier transforms.
        for i in range(n // 2):
            apply_phase(register, start + 2 * i + 1, -i * numpy.pi / n)
            fourier_transform_0_adjacent(register, start + 2 * i)

        # Undo the swap network.
        for i in range(n//2 - 1, -1, -1):
            for j in range(i):
                # fswap is equivalent to Swap followed by C(Z)
                fswap_adjacent(register, center + 2*j - i)


def ffft_2d(engine, register, system_size):
    """Apply the 2D fermionic fast Fourier transform to a register.

    Args:
        engine (projectq.MainEngine): The simulator engine with the
                                      register.
        register (projectq.QuReg): The register to apply the 2D FFFT to.
        system_size (int): The side length of the system. register must
                           thus have system_size ** 2 qubits.

    Notes:
        This algorithm uses radix-2 decimation-in-time, so system_size
        must be a binary power. This decimation is head recursive (the
        2-mode FFFT is applied as the first part of the 4-mode FFFT,
        which is applied as the first part of the 8-mode FFFT, etc).
    """
    # Apply the FFFT along one axis of the register.
    for i in range(system_size):
        ffft(engine, register[system_size*i: system_size*i + system_size],
             system_size)
        Ph(3 * numpy.pi / 4) | register[0]

    # To apply the FFFT along the second axis, we must fermionically
    # swap qubits into the correct positions. In 2D this is equivalent
    # to flipping all qubits across the "diagonal" of the grid.
    key_2d = (index_of_position_in_1d_array, (0, 1))
    arr = [(i, j) for i in range(system_size)
           for j in range(system_size)]
    swaps = parallel_bubble_sort(arr, key_2d, system_size)
    all_swaps = [swap for swap_layer in swaps for swap in swap_layer]

    # Fermionically swap into position to FFFT along the second axis.
    for swap in all_swaps:
        Swap | (register[swap[0]], register[swap[1]])
        C(Z) | (register[swap[0]], register[swap[1]])

    # Perform the FFFT along the second axis.
    for i in range(system_size):
        ffft(engine, register[system_size*i: system_size*i + system_size],
             system_size)
        Ph(3 * numpy.pi / 4) | register[0]

    # Undo the fermionic swap network to restore the original ordering.
    for swap in all_swaps[::-1]:
        Swap | (register[swap[0]], register[swap[1]])
        C(Z) | (register[swap[0]], register[swap[1]])


def swap_adjacent_fermionic_modes(fermion_operator, mode):
    """Swap adjacent modes in a fermionic operator.

    Returns: a new FermionOperator with mode and mode+1 swapped.

    Args:
        fermion_operator (projectq.FermionOperator): Original operator.
        mode (integer): The mode to be swapped with mode + 1.

    Notes:
        Because the swap must be fermionic, the sign of the operator is
        flipped if both creation operators mode^ and (mode+1)^ (or the
        corresponding annihilation operators) are present.
    """
    new_operator = FermionOperator.zero()

    for term in fermion_operator.terms:
        new_term = list(term)
        multiplier = 1
        if (mode, 1) in term and (mode + 1, 1) in term:
            multiplier *= -1
        else:
            if (mode, 1) in term:
                new_term[term.index((mode, 1))] = (mode + 1, 1)
            if (mode + 1, 1) in term:
                new_term[term.index((mode + 1, 1))] = (mode, 1)

        if (mode, 0) in term and (mode + 1, 0) in term:
            multiplier *= -1
        else:
            if (mode, 0) in term:
                new_term[term.index((mode, 0))] = (mode + 1, 0)
            if (mode + 1, 0) in term:
                new_term[term.index((mode + 1, 0))] = (mode, 0)

        new_operator.terms[tuple(new_term)] = (multiplier *
                                               fermion_operator.terms[term])

    return new_operator


def ffft_swap_networks(system_size, n_dimensions):
    """Compute the two networks of swaps required for the FFFT.

    Args:
        system_size (int): The system side length (the system is assumed
                           to be a hypercube).
        n_dimensions (int): Number of dimensions for the system.

    Notes:
        The reordering of the operators is the opposite of the ordering
        that would be applied to a state. This reordering is given in
        two stages. First, the bit-reversed output of the FFFT circuit
        (e.g. 0, 2, 1, 3 which is the four numbers 0, 1, 2, 3 after
        reversing their bit representations) is swapped to the standard
        order (e.g. 0, 1, 2, 3). Second, the standard order is swapped
        to the split-halves ordering used in fourier_transform (e.g.
        2, 3, 0, 1 which swaps the left and right halves of that range).

        Both rounds must be applied before the FFFT, but after the FFFT
        only the second round needs to be undone so as to restore to the
        order of the FFFT circuit. All reorderings must be done in both
        dimensions, and because the operator is fermionic all swaps must
        also be.
    """
    # Create a string formatting rule for converting a number into binary
    # using log2(system_size) digits.
    binary_format_rule = '{:0' + str(int(numpy.log2(system_size))) + 'b}'

    # Use the formatting rule to generate the bit-reversed output of the FFFT.
    bit_reversed_order = [int(binary_format_rule.format(i)[::-1], 2)
                          for i in range(system_size)]

    # Generate the split-halves order of fourier_transform.
    split_halves_order = (list(range(system_size // 2, system_size)) +
                          list(range(system_size // 2)))

    # Enumerate system coordinates.
    coordinates = list(itertools.product(*[range(system_size)
                                           for i in range(n_dimensions)]))

    # Create arrays of the grid coordinates corresponding to the two orders.
    # Within each dimension, the grid coordinates are in either bit-reversed
    # or split-halves ordering.
    bit_reversed_order_coordinates = [None] * system_size ** n_dimensions
    split_halves_order_coordinates = [None] * system_size ** n_dimensions
    for i in range(system_size ** n_dimensions):
        # e.g. for 4x4, this will be [(0, 0), (0, 2), (0, 1), (0, 3), (2, 0),
        # (2, 2), (2, 1), (2, 3), (1, 0), ...]; it is ordered along both dims
        # by the bit-reversed order 0, 2, 1, 3. The split-halves coordinates
        # are constructed similarly.
        bit_reversed_order_coordinates[i] = tuple(map(
            bit_reversed_order.__getitem__, coordinates[i]))
        split_halves_order_coordinates[i] = tuple(map(
            split_halves_order.__getitem__, coordinates[i]))

    # Find the two rounds of swaps for parallel bubble sort into
    # row-major order.
    key = (index_of_position_in_1d_array, range(n_dimensions - 1, -1, -1))
    first_round_swaps = parallel_bubble_sort(bit_reversed_order_coordinates,
                                             key, system_size)
    second_round_swaps = (
        parallel_bubble_sort(split_halves_order_coordinates,
                             key, system_size)[::-1])

    return [first_round_swaps, second_round_swaps]


def operator_2d_fft_with_reordering(fermion_operator, system_size):
    """Apply the 2D FFT to an operator after reordering its modes.

    Args:
        fermion_operator (projectq.FermionOperator): Original operator.
        system_size (int): The side length of the system. register must
                           thus have system_size ** 2 qubits.

    Notes:
        The reordering of the operators is the opposite of the ordering
        that would be applied to a state. This reordering is given in
        two stages. First, the bit-reversed output of the FFFT circuit
        (e.g. 0, 2, 1, 3 which is the four numbers 0, 1, 2, 3 after
        reversing their bit representations) is swapped to the standard
        order (e.g. 0, 1, 2, 3). Second, the standard order is swapped
        to the split-halves ordering used in fourier_transform (e.g.
        2, 3, 0, 1 which swaps the left and right halves of that range).

        Both rounds must be applied before the FFFT, but after the FFFT
        only the second round needs to be undone so as to restore to the
        order of the FFFT circuit. All reorderings must be done in both
        dimensions, and because the operator is fermionic all swaps must
        also be.
    """
    fermion_operator = normal_ordered(fermion_operator)
    grid = Grid(dimensions=2, length=system_size, scale=1.0)

    first_round_swaps, second_round_swaps = ffft_swap_networks(system_size,
                                                               n_dimensions=2)

    # Create the full list of swaps.
    swap_mode_list = first_round_swaps + second_round_swaps
    swaps = [swap[0] for swap_layer in swap_mode_list for swap in swap_layer]

    # Swap adjacent modes for all swaps in both rounds.
    for mode in swaps:
        fermion_operator = swap_adjacent_fermionic_modes(
            fermion_operator, mode)
        fermion_operator = normal_ordered(fermion_operator)

    # Apply the Fourier transform to the reordered operator.
    fft_result = fourier_transform(fermion_operator, grid, spinless=True)
    fft_result = normal_ordered(fft_result)

    # Undo the second round of swaps to restore the FFT's bit-reversed order.
    swap_mode_list = second_round_swaps
    swaps = [swap[0] for swap_layer in swap_mode_list for swap in swap_layer]
    for mode in swaps:
        fft_result = swap_adjacent_fermionic_modes(fft_result, mode)
        fft_result = normal_ordered(fft_result)

    return fft_result



"""Module for low-depth Trotterized simulation in the plane wave dual basis."""
import functools
import itertools
import numpy

from openfermionprojectq._parallel_bubble_sort import (
    index_of_position_in_1d_array, parallel_bubble_sort)

from openfermion.ops import FermionOperator
from openfermion.transforms import jordan_wigner
from openfermion.utils import count_qubits

from projectq.ops import (C, S, Sdag, X, Y, Z, H, CNOT, Swap, Rx, Ry, Rz, Ph,
                          All, Measure)


def fermionic_reorder(register, input_order, target_order=None):
    """Fermionically reorder the given register.

    Maps the input Jordan-Wigner order to the output order.

    Args:
        register (projectq.QuReg): Quantum register to reorder.
        input_order (list): The initial Jordan-Wigner canonical order.
        target_order (list): The desired Jordan-Wigner canonical order.
    """
    if not target_order:
        target_order = list(range(len(input_order)))

    key = (index_of_position_in_1d_array, [0])

    input_swaps = parallel_bubble_sort(list(itertools.product(input_order)),
                                       key, len(input_order))
    input_swaps = [swap[0] for swap_layer in input_swaps
                   for swap in swap_layer]

    output_swaps = parallel_bubble_sort(list(itertools.product(target_order)),
                                        key, len(input_order))
    output_swaps = [swap[0] for swap_layer in output_swaps
                    for swap in swap_layer]

    # Invert the output swaps to go from the input to the target.
    swaps = input_swaps + output_swaps[::-1]

    # Swap adjacent modes for all swaps in both rounds.
    for mode in swaps:
        C(Z) | (register[mode], register[mode + 1])
        Swap | (register[mode], register[mode + 1])


def special_F_adjacent(register, qubit_index, xx_yy_angle, zz_angle):
    """Apply the 'special F' (fermionic swap, XX+YY evolution, and ZZ
    evolution) gate to the register on qubit_index and qubit_index + 1.

    Args:
        register (projectq.QuReg): The register to apply the gate to.
        qubit_index (integer): The left qubit to act on.
        xx_yy_angle (float): The angle for evolution under XX+YY.
        zz_angle (float): The angle for evolution under ZZ.
    """
    Rx(numpy.pi / 2.) | register[qubit_index]
    CNOT | (register[qubit_index], register[qubit_index + 1])
    if numpy.abs(xx_yy_angle) > 0.0:
        Rx(xx_yy_angle) | register[qubit_index]
        Ry(xx_yy_angle) | register[qubit_index + 1]
    CNOT | (register[qubit_index + 1], register[qubit_index])
    Rx(-numpy.pi / 2.) | register[qubit_index + 1]
    if numpy.abs(zz_angle) > 0.0:
        Rz(zz_angle) | register[qubit_index + 1]
    Sdag | register[qubit_index + 1]
    CNOT | (register[qubit_index], register[qubit_index + 1])
    S | register[qubit_index]
    S | register[qubit_index + 1]


def simulation_gate_trotter_step(register, hamiltonian, input_ordering=None,
                                 first_order=True):
    """Simulate a unit time Trotter step under the plane wave dual basis
    Hamiltonian using the fermionic simulation gate.

    Args:
        register (projectq.QuReg): The register to apply the unitary to.
        hamiltonian (QubitOperator): Qubit operator representing the problem
            in the plane-wave dual basis.
        input_ordering (list): The input Jordan-Wigner ordering.
        first_order (bool): Whether to apply a first or second-order
                            Trotter step.

    Notes:
        Applying a first-order Trotter step reverses the input ordering.
    """
    n_qubits = count_qubits(hamiltonian)

    if not input_ordering:
        input_ordering = list(range(n_qubits))

    # The intermediate ordering is the halfway point for second-order Trotter.
    intermediate_ordering = input_ordering[::-1]
    if first_order:
        final_ordering = intermediate_ordering
    else:
        final_ordering = input_ordering[:]

    # Whether we're in an odd or an even stagger. Alternates at each step.
    odd = 0

    # If we do a second order Trotter step, the input and final orderings
    # are the same. Use a flag to force us to leave the initial step.
    moved = False

    while input_ordering != final_ordering or not moved:
        moved = True
        for i in range(odd, n_qubits - 1, 2):
            left = input_ordering[i]
            right = input_ordering[i + 1]
            # For real c, c*([3^ 2] + [2^ 3]) transforms under JW to
            # c/2 (X_2 X_3 + Y_2 Y_3); evolution is exp(-ic/2(XX+YY))
            xx_yy_angle = hamiltonian.terms.get(((left, 1), (right, 0)), 0.0)

            left, right = max(left, right), min(left, right)
            # Two-body terms c 2^ 1^ 2 1 are equal to -c n_2 n_1. JW maps this
            # to -c(I-Z_2)(I-Z_1)/4 = -c I/4 + c Z_1/4 + c Z_2/4 - c Z_1 Z_2/4.
            # Evolution is thus exp(ic/4*(I - Z_1 - Z_2 + Z_1 Z_2)).
            zz_angle = -hamiltonian.terms.get(
                ((left, 1), (right, 1), (left, 0), (right, 0)), 0.0) / 2.

            # Divide by two for second order Trotter.
            if not first_order:
                xx_yy_angle /= 2
                zz_angle /= 2

            special_F_adjacent(register, i, xx_yy_angle, zz_angle)
            # The single-Z rotation angle is the opposite of the ZZ angle.
            if numpy.abs(zz_angle) > 0.0:
                Rz(-zz_angle) | register[i]
                Rz(-zz_angle) | register[i + 1]
                Ph(-zz_angle / 2.) | register[0]

            num_operator_left = ((input_ordering[i], 1),
                                 (input_ordering[i], 0))
            if num_operator_left in hamiltonian.terms:
                # Jordan-Wigner maps a number term c*n_i to c*(I-Z_i)/2.
                # Time evolution is then exp(-i c*(I-Z_i)/2), which is equal to
                # a phase exp(-ic/2) and a rotation Rz(-c).
                z_angle = (-hamiltonian.terms[num_operator_left] /
                           (n_qubits - 1))

                # Divide by two for second order Trotter.
                if not first_order:
                    z_angle /= 2.
                # After special_F_adjacent, qubits i and i+1 have swapped;
                # the ith rotation must thus be applied to qubit i+1.
                if numpy.abs(z_angle) > 0.0:
                    Rz(z_angle) | register[i + 1]
                    Ph(z_angle / 2.) | register[0]

            num_operator_right = ((input_ordering[i+1], 1),
                                  (input_ordering[i+1], 0))
            if num_operator_right in hamiltonian.terms:
                # Jordan-Wigner maps a number term c*n_i to c*(I-Z_i)/2.
                # Time evolution is then exp(-i c*(I-Z_i)/2), which is equal to
                # a phase exp(-ic/2) and a rotation Rz(-c).
                z_angle = (-hamiltonian.terms[num_operator_right] /
                           (n_qubits - 1))

                # Divide by two for second order Trotter.
                if not first_order:
                    z_angle /= 2
                # After special_F_adjacent, qubits i and i+1 have swapped;
                # the (i+1)th rotation must thus be applied to qubit i.
                if numpy.abs(z_angle) > 0.0:
                    Rz(z_angle) | register[i]
                    Ph(z_angle / 2.) | register[0]

            # Finally, swap the two modes in input_ordering.
            input_ordering[i], input_ordering[i + 1] = (input_ordering[i + 1],
                                                        input_ordering[i])

        # Unless we're at the intermediate ordering, odd should flip.
        # This is only needed for second-order Trotter.
        if input_ordering != intermediate_ordering:
            odd = 1 - odd

    return input_ordering


def simulate_dual_basis_evolution(register, hamiltonian, trotter_steps=1,
                                  input_ordering=None, first_order=True):
    """Simulate Trotterized evolution under the plane wave Hamiltonian
    in the plane wave dual basis.

    Args:
        register (projectq.QuReg): The register to apply the unitary to.
        hamiltonian (FermionOperator): The normal-ordered dual basis
                                       Hamiltonian to simulate.
        trotter_steps (int): The number of Trotter steps to divide into.
        input_ordering (list): The input Jordan-Wigner ordering.
        first_order (bool): Whether to apply first or second-order
                            Trotter steps.

    Notes:
        Applying an odd number of first-order Trotter steps reverses
        the input ordering.
    """
    n_qubits = count_qubits(hamiltonian)

    if not input_ordering:
        input_ordering = list(range(n_qubits))

    if set(input_ordering) != set(range(n_qubits)):
        raise ValueError('input_ordering must be a permutation of integers '
                         '0 through the number of qubits.')

    if not isinstance(trotter_steps, int) or trotter_steps < 1:
        raise ValueError('The number of Trotter steps must be an int >0.')

    trotterized_hamiltonian = hamiltonian / float(trotter_steps)

    for i in range(trotter_steps):
        input_ordering = simulation_gate_trotter_step(
            register, trotterized_hamiltonian, input_ordering, first_order)

    return input_ordering



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to test unitary coupled cluster operators."""

from __future__ import absolute_import

from numpy.random import randn
import scipy

import itertools
import os
import unittest

import openfermion
from openfermion.config import THIS_DIRECTORY
from openfermion.hamiltonians import MolecularData
from openfermion.ops import *
from openfermion.transforms import *
from openfermion.utils import *

from openfermionprojectq import TimeEvolution
from openfermionprojectq._graph import (Graph, Node)
from openfermionprojectq._unitary_cc import *

from projectq.ops import All, Measure, X


class UnitaryCC(unittest.TestCase):

    def test_simulation_energy(self):
        """Test UCCSD Singlet Energy for H2"""

        # Define H2 Hamiltonian inline
        hamiltonian = ((-0.0453222020986) * QubitOperator("X0 X1 Y2 Y3") +
                       (0.165867023964) * QubitOperator("Z0 Z3") +
                       (0.174348441706) * QubitOperator("Z2 Z3") +
                       (0.120544821866) * QubitOperator("Z0 Z2") +
                       (3.46944695195e-18) * QubitOperator("X0 Y1 X2 Y3") +
                       (0.165867023964) * QubitOperator("Z1 Z2") +
                       (0.171197748533) * QubitOperator("Z0") +
                       (-0.222785928901) * QubitOperator("Z3") +
                       (3.46944695195e-18) * QubitOperator("X0 X1 X2 X3") +
                       (0.168622191433) * QubitOperator("Z0 Z1") +
                       (0.120544821866) * QubitOperator("Z1 Z3") +
                       (3.46944695195e-18) * QubitOperator("Y0 Y1 Y2 Y3") +
                       (-0.0988639735178) * QubitOperator("") +
                       (0.171197748533) * QubitOperator("Z1") +
                       (0.0453222020986) * QubitOperator("Y0 X1 X2 Y3") +
                       (3.46944695195e-18) * QubitOperator("Y0 X1 Y2 X3") +
                       (-0.0453222020986) * QubitOperator("Y0 Y1 X2 X3") +
                       (-0.222785928901) * QubitOperator("Z2") +
                       (0.0453222020986) * QubitOperator("X0 Y1 Y2 X3"))
        hamiltonian.compress()
        compiler_engine = uccsd_trotter_engine()
        wavefunction = compiler_engine.allocate_qureg(4)
        test_amplitudes = [-1.14941450e-08, 5.65340614e-02]
        for i in range(2):
            X | wavefunction[i]
        evolution_operator = uccsd_singlet_evolution(test_amplitudes, 4, 2)
        evolution_operator | wavefunction
        compiler_engine.flush()
        energy = compiler_engine.backend.get_expectation_value(hamiltonian,
                                                               wavefunction)
        All(Measure) | wavefunction
        self.assertAlmostEqual(energy, -1.13727017463)

    def test_simulation_with_graph(self):
        """Test UCCSD Singlet Energy for H2 using a restricted qubit_graph"""

        # Define H2 Hamiltonian inline
        hamiltonian = ((-0.0453222020986) * QubitOperator("X0 X1 Y2 Y3") +
                       (0.165867023964) * QubitOperator("Z0 Z3") +
                       (0.174348441706) * QubitOperator("Z2 Z3") +
                       (0.120544821866) * QubitOperator("Z0 Z2") +
                       (3.46944695195e-18) * QubitOperator("X0 Y1 X2 Y3") +
                       (0.165867023964) * QubitOperator("Z1 Z2") +
                       (0.171197748533) * QubitOperator("Z0") +
                       (-0.222785928901) * QubitOperator("Z3") +
                       (3.46944695195e-18) * QubitOperator("X0 X1 X2 X3") +
                       (0.168622191433) * QubitOperator("Z0 Z1") +
                       (0.120544821866) * QubitOperator("Z1 Z3") +
                       (3.46944695195e-18) * QubitOperator("Y0 Y1 Y2 Y3") +
                       (-0.0988639735178) * QubitOperator("") +
                       (0.171197748533) * QubitOperator("Z1") +
                       (0.0453222020986) * QubitOperator("Y0 X1 X2 Y3") +
                       (3.46944695195e-18) * QubitOperator("Y0 X1 Y2 X3") +
                       (-0.0453222020986) * QubitOperator("Y0 Y1 X2 X3") +
                       (-0.222785928901) * QubitOperator("Z2") +
                       (0.0453222020986) * QubitOperator("X0 Y1 Y2 X3"))
        hamiltonian.compress()

        # Create a star graph of 4 qubits, all connected through qubit 0
        qubit_graph = Graph()
        compiler_engine = uccsd_trotter_engine(qubit_graph=qubit_graph,
                                               opt_size=3)
        wavefunction = compiler_engine.allocate_qureg(4)
        for i in range(4):
            qubit_graph.add_node(Node(value=wavefunction[i].id))
        for i in range(1, 4):
            qubit_graph.add_edge(0, i)

        test_amplitudes = [-1.14941450e-08, 5.65340614e-02]
        for i in range(2):
            X | wavefunction[i]
        evolution_operator = uccsd_singlet_evolution(test_amplitudes, 4, 2)
        evolution_operator | wavefunction
        compiler_engine.flush()

        energy = compiler_engine.backend.get_expectation_value(hamiltonian,
                                                               wavefunction)
        self.assertAlmostEqual(energy, -1.13727017463)

        # Check swap only with non-adjacent qubits functions without error
        projectq.ops.Swap | (wavefunction[0], wavefunction[3])
        projectq.ops.Swap | (wavefunction[0], wavefunction[3])
        compiler_engine.flush()
        energy = compiler_engine.backend.get_expectation_value(hamiltonian,
                                                               wavefunction)
        All(Measure) | wavefunction
        self.assertAlmostEqual(energy, -1.13727017463)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""This module has functions to perform parallel bubble sort on arrays.

Within the context of openfermion, parallel bubble sort can be used to
quickly reorder fermionic modes or qubits.
"""
from functools import partial


def index_of_position_in_1d_array(coordinate_ordering, system_side_length,
                                  position):
    """Return the index of an n-dimensional position embedded in a
    1-dimensional array.

    The index is computed using the given ordering of coordinate
    weights as well as the side length of the system from which the
    n-dimensional position is taken.

    Args:
        coordinate_ordering (list of ints): The order of coordinates by
                which to increment the index.
        system_side_length (int): The number of points along a side of
                the system.
        position (list of ints): The vector indicating the position
                within the n-dimensional system.

    Returns:
        The integer index of the position in a 1D array.

    Examples:
        index_of_position_in_1d_array((2, 1, 0), 4, (1, 2, 3))
                -> 27 (16 + 4 * 2 + 1 * 3).
        This is standard lexicographic ordering for 3D positions.

        index_of_position_in_1d_array((0, 1, 2), 2, (0, 1, 1))
                -> 6 (2 * 1 + 4 * 1)

        index_of_position_in_1d_array((2, 0, 1), 3, (2, 1, 0))
                -> 19 (9 * 2 + 1 + 3 * 0)
    """
    return sum(position[i] * system_side_length ** coordinate_ordering[i]
               for i in range(len(coordinate_ordering)))


def is_sorted_array_of_nd_positions(arr, key, system_side_length):
    """Determine whether an array of n-dim positions is sorted.

    Args:
        arr: the array to work with.
        key: the key function or tuple to use to determine ordering.
        system_side_length: The integer side length of the n-dimensional
                            cube from which the positions in the array
                            are taken.
    """
    if isinstance(key, tuple):
        key = partial(*key)

    for i in range(len(arr) - 1):
        if (key(system_side_length, arr[i]) >
                key(system_side_length, arr[i + 1])):
            return False
    return True


def parallel_bubble_sort(array, key, system_side_length):
    """Give the layers of swaps which sort an array of positions.

    Swaps can occur only between adjacent entries in the array.

    Args:
        array: An array of multi-dimensional positions to sort.
        key: The key function or tuple by which to determine ordering.
        system_side_length: The integer side length of the hypercube
                            from which the positions in array are taken.
    """
    swaps = []
    odd = 0

    if isinstance(key, tuple):
        key = partial(*key)

    while not is_sorted_array_of_nd_positions(array, key, system_side_length):
        swaps.append(parallel_bubble_sort_single_step(
            array, key, odd, system_side_length=system_side_length))
        odd = 1 - odd

    return swaps


def parallel_bubble_sort_single_step(array, key, odd=False,
                                     system_side_length=None):
    """Give a layer of swaps as part of a sorting network for an array of
    positions.

    Args:
        array: An array of multi-dimensional positions to sort.
        key: The key function by which to determine ordering.
        odd: Whether or not to swap all odd entries with their even
             neighbor (whether to the right or down), e.g. swapping
             x[i][j] with x[i+1][j] or x[i][j+1] with i or j even, which
             happens when odd = False, or with i or j odd, which happens
             when odd = True.
        system_side_length: The integer side length of the n-cube from
                            which the positions in array are taken.
    """
    swaps_in_layer = []

    for i in range(int(odd), len(array) - 1, 2):
        if (key(system_side_length, array[i]) >
                key(system_side_length, array[i + 1])):
            array[i], array[i+1] = array[i+1], array[i]
            swaps_in_layer.append((i, i + 1))

    return swaps_in_layer



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to create and manipulate unitary coupled cluster operators."""

from openfermionprojectq import TimeEvolution
from openfermion.ops import FermionOperator, QubitOperator
from openfermion.transforms import jordan_wigner
from openfermion.utils import (uccsd_generator, uccsd_singlet_generator,
                               uccsd_singlet_paramsize)

import projectq
import projectq.setups
import projectq.setups.decompositions


def uccsd_evolution(fermion_generator, fermion_transform=jordan_wigner):
    """Create a ProjectQ evolution operator for a UCCSD circuit

    Args:
        fermion_generator(FermionOperator): UCCSD generator to evolve.
        fermion_transform(openfermion.transform): The transformation that
            defines the mapping from Fermions to QubitOperator.

    Returns:
        evoution_operator(TimeEvolution): The unitary operator
            that constructs the UCCSD state.
    """

    # Transform generator to qubits
    qubit_generator = fermion_transform(fermion_generator)

    # Cast to real part only for compatibility with current ProjectQ routine
    for key in qubit_generator.terms:
        qubit_generator.terms[key] = float(qubit_generator.terms[key].imag)
    qubit_generator.compress()

    # Allocate wavefunction and act evolution on gate according to compilation
    evolution_operator = (
        TimeEvolution(time=1., hamiltonian=qubit_generator))

    return evolution_operator


def uccsd_singlet_evolution(packed_amplitudes, n_qubits, n_electrons,
                            fermion_transform=jordan_wigner):
    """Create a ProjectQ evolution operator for a UCCSD singlet circuit

    Args:
        packed_amplitudes(ndarray): Compact array storing the unique single
            and double excitation amplitudes for a singlet UCCSD operator.
            The ordering lists unique single excitations before double
            excitations.
        n_qubits(int): Number of spin-orbitals used to represent the system,
            which also corresponds to number of qubits in a non-compact map.
        n_electrons(int): Number of electrons in the physical system
        fermion_transform(openfermion.transform): The transformation that
            defines the mapping from Fermions to QubitOperator.

    Returns:
        evoution_operator(TimeEvolution): The unitary operator
            that constructs the UCCSD singlet state.
    """
    # Build UCCSD generator
    fermion_generator = uccsd_singlet_generator(packed_amplitudes,
                                                n_qubits,
                                                n_electrons)

    evolution_operator = uccsd_evolution(fermion_generator,
                                         fermion_transform)

    return evolution_operator


def _identify_non_commuting(cmd):
    """Recognize all TimeEvolution gates with >1 terms that don't all commute.

    This is a filter function for use with ProjectQ that flags terms as
        non-commuting so they may be handled by a different factorization
        routine.

    Args:
        cmd(projectq.command): A command from ProjectQ

    Returns:
        (bool) Depending on whether the terms are determined to commute or not
    """
    hamiltonian = cmd.gate.hamiltonian
    if len(hamiltonian.terms) == 1:
        return False
    else:
        id_op = QubitOperator((), 0.0)
        for term in hamiltonian.terms:
            test_op = QubitOperator(term, hamiltonian.terms[term])
            for other in hamiltonian.terms:
                other_op = (QubitOperator(other, hamiltonian.terms[other]))
                commutator = test_op * other_op - other_op * test_op
                if not commutator == id_op:
                    return True
    return False


def _non_adjacent_filter(self, cmd, qubit_graph, flip=False):
    """A ProjectQ filter to identify when swaps are needed on a graph

    This flags any gates that act on two non-adjacent qubits with respect to
    the qubit_graph that has been given

    Args:
        self(Dummy): Dummy parameter to meet function specification.
        cmd(projectq.command): Command to be checked for decomposition into
            additional swap gates.
        qubit_graph(Graph): Graph object specifying connectivity of
            qubits. The values of the nodes of this graph are unique qubit ids.
        flip(Bool): Flip for switching if identifying a gate is in this class
            by true or false.  Designed to meet the specification of ProjectQ
            InstructionFilter and DecompositionRule with one function.

    Returns:
        bool: When flip is False, this returns True when a 2 qubit command
            acts on non-adjacent qubits or when it acts only on a single qubit.
            This is reversed when flip is used.

    """
    if qubit_graph is None:
        return True ^ flip

    total_qubits = (cmd.control_qubits +
                    [item for qureg in cmd.qubits for item in qureg])

    # Check for non-connected gate on 2 qubits
    if ((len(total_qubits) == 1) or
            (len(total_qubits) == 2 and
             qubit_graph.is_adjacent(
                 qubit_graph.find_index(total_qubits[0].id),
                 qubit_graph.find_index(total_qubits[1].id)))):
        return True ^ flip
    return False ^ flip


def _direct_graph_swap(cmd, qubit_graph):
    """Define a naive direct swap sequence to respect qubit_graph connectivity

    Uses the connectivity of qubit_graph to find the shortest path between
    two non-adjacent qubits, and swaps/unswaps qubits appropriately.  Baseline
    for more sophisticated algorithms

    Args:
        cmd(projectq.command): A command from ProjectQ that needs to be
            broken down due to non-adjacent terms
        qubit_graph(Graph): Graph object specifying connectivity of qubits.
            The values of the nodes of this graph are unique qubit ids
    """
    total_qubits = (cmd.control_qubits +
                    [item for qureg in cmd.qubits for item in qureg])

    gate = cmd.gate
    engine = cmd.engine
    graph_path = qubit_graph.shortest_path(
        qubit_graph.find_index(total_qubits[0].id),
        qubit_graph.find_index(total_qubits[1].id))
    swap_path = [(graph_path[i], graph_path[i + 1])
                 for i in range(len(graph_path) - 2)]

    # SWAP qubit 1 into position adjacent to qubit 2
    for pair in swap_path:
        projectq.ops.Swap | (projectq.types.
                             WeakQubitRef(engine,
                                          qubit_graph.nodes[pair[0]].value),
                             projectq.types.
                             WeakQubitRef(engine,
                                          qubit_graph.nodes[pair[1]].value))

    # Perform original gate - assumes use of standard ProjectQ rule set that
    # only ends with 2-qubit gates with controls. Swaps decomposed to CX
    projectq.ops.C(gate) | (projectq.types.
                            WeakQubitRef(engine,
                                         qubit_graph.
                                         nodes[graph_path[-2]].value),
                            total_qubits[1])

    # Reverse the swaps to put qubits back in place
    for pair in reversed(swap_path):
        projectq.ops.Swap | (projectq.types.
                             WeakQubitRef(engine,
                                          qubit_graph.nodes[pair[0]].value),
                             projectq.types.
                             WeakQubitRef(engine,
                                          qubit_graph.nodes[pair[1]].value))


def _first_order_trotter(cmd):
    """Define a Trotter splitting for non-commuting Pauli in ProjectQ

    This routine defines a first-order Trotter splitting to be applied to
    time evolution operators in ProjectQ.

    Args:
        cmd(projectq.command): A command from ProjectQ that needs to be
            factorized due to non-commuting time evolution terms
    """
    qureg = cmd.qubits
    eng = cmd.engine
    hamiltonian = cmd.gate.hamiltonian
    time = cmd.gate.time
    with projectq.meta.Control(eng, cmd.control_qubits):
        # First order Trotter splitting
            for term in hamiltonian.terms:
                ind_operator = (QubitOperator(term, hamiltonian.terms[term]))
                TimeEvolution(time, ind_operator) | qureg


def _two_gate_filter(self, cmd):
    """A ProjectQ filter to flag TimeEvolution operators for decomposition

    This flags any gates which act on more than 2 qubits or are time evolution
    operators to be decomposed into a base library of gates for simulation
    within ProjectQ.

    Args:
        self(Dummy): Dummy parameter to meet function specification.
        cmd(projectq.command): Command to be checked for decomposition into
            one- and two- qubit gates.

    """
    if ((not isinstance(cmd.gate, projectq.ops.TimeEvolution)) and
        (len(cmd.qubits[0]) <= 2 or
            isinstance(cmd.gate, projectq.ops.ClassicalInstructionGate))):
        return True
    return False


def uccsd_trotter_engine(compiler_backend=projectq.backends.Simulator(),
                         qubit_graph=None, opt_size=None):
    """Define a ProjectQ compiler engine that is common for use with UCCSD

    This defines a ProjectQ compiler engine that decomposes time evolution
    gates using a first order Trotter decomposition on non-commuting gates
    down to a base gate decomposition.

    Args:
        compiler_backend(projectq.backend): Define the backend on the
            circuit compiler, so that it may either simulate gates numerically
            or alternatively print a gate sequence, e.g. using
            projectq.backends.CommandPrinter()
        qubit_graph(Graph): Graph object specifying connectivity of qubits.
            The values of the nodes of this unique qubit ids.  If None,
            all-to-all connectivity is assumed.
        opt_size(int): Number for ProjectQ local optimizer to determine size
            of blocks optimized over.

    Returns:
        projectq.cengine that is the compiler engine set up with these
            rules and decompostions.
    """
    rule_set = (
        projectq.cengines.
        DecompositionRuleSet(modules=[projectq.setups.decompositions]))

    # Set rules for splitting non-commuting operators
    trotter_rule_set = (projectq.cengines.DecompositionRule(
        gate_class=TimeEvolution,
        gate_decomposer=_first_order_trotter,
        gate_recognizer=_identify_non_commuting))
    rule_set.add_decomposition_rule(trotter_rule_set)

    # Set rules for 2 qubit gates that act on non-adjacent qubits
    if qubit_graph is not None:
        connectivity_rule_set = (
            projectq.cengines.DecompositionRule(
                gate_class=projectq.ops.NOT.__class__,
                gate_decomposer=(lambda x: _direct_graph_swap(x, qubit_graph)),
                gate_recognizer=(lambda x: _non_adjacent_filter(None, x,
                                                                qubit_graph,
                                                                True))))
        rule_set.add_decomposition_rule(connectivity_rule_set)

    # Build the full set of engines that will be applied to qubits
    replacer = projectq.cengines.AutoReplacer(rule_set)
    compiler_engine_list = [replacer,
                            projectq.
                            cengines.
                            InstructionFilter(
                                lambda x, y:
                                (_non_adjacent_filter(x, y, qubit_graph) and
                                 _two_gate_filter(x, y)))]
    if opt_size is not None:
        compiler_engine_list.append(projectq.cengines.LocalOptimizer(opt_size))

    # Start the compiler engine with these rules
    compiler_engine = (
        projectq.MainEngine(backend=compiler_backend,
                            engine_list=compiler_engine_list))
    return compiler_engine



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os

from setuptools import setup, find_packages

# This reads the __version__ variable from openfermionprojectq/_version.py
exec(open('openfermionprojectq/_version.py').read())

# Readme file as long_description:
long_description = open('README.rst').read()

# Read in requirements.txt
requirements = open('requirements.txt').readlines()
requirements = [r.strip() for r in requirements]

setup(
    name='openfermionprojectq',
    version=__version__,
    author='The OpenFermion Developers',
    author_email='help@openfermion.org',
    url='http://www.openfermion.org',
    description=('A plugin allowing OpenFermion to interaface with ProjectQ.'),
    long_description=long_description,
    install_requires=requirements,
    license='Apache 2',
    packages=find_packages()
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for _parallel_bubble_sort.py."""
import unittest

from openfermionprojectq._parallel_bubble_sort import (
    index_of_position_in_1d_array, is_sorted_array_of_nd_positions,
    parallel_bubble_sort)


class IndexTest(unittest.TestCase):
    def test_index_standard_order_coordinates(self):
        self.assertEqual(
            index_of_position_in_1d_array(
                (2, 1, 0), 3, (1, 2, 3)),
            ((3 ** 2) * 1 + (3 ** 1) * 2 + (3 ** 0) * 3))

    def test_index_reverse_order_coordinates(self):
        self.assertEqual(
            index_of_position_in_1d_array(
                (0, 1, 2), 3, (2, 3, 5)),
            ((3 ** 0) * 2 + (3 ** 1) * 3 + (3 ** 2) * 5))

    def test_index_scrambled_coordinates(self):
        self.assertEqual(
            index_of_position_in_1d_array(
                (2, 0, 1), 4, (2, 1, 0)),
            ((4 ** 2) * 2 + (4 ** 0) * 1 + (4 ** 1) * 0))

    def test_index_scrambled_coordinates_6dimensions(self):
        self.assertEqual(
            index_of_position_in_1d_array(
                (1, 2, 0, 5, 4, 3), 3, (2, 1, 0, 0, 1, 2)),
            (3 ** 1) * 2 + (3 ** 2) + (3 ** 4) + (3 ** 3) * 2)


class IsSortedTest(unittest.TestCase):
    def test_is_sorted_standard_order(self):
        key = (index_of_position_in_1d_array, (2, 1, 0))
        arr = [(0, 0, i) for i in range(7)]
        self.assertTrue(is_sorted_array_of_nd_positions(arr, key, 7))

    def test_not_is_sorted_reverse_ordered_list(self):
        key = (index_of_position_in_1d_array, (2, 1, 0))
        arr = [(0, 0, i) for i in range(7, -1, -1)]
        self.assertFalse(is_sorted_array_of_nd_positions(arr, key, 7))

    def test_is_sorted_3d(self):
        key = (index_of_position_in_1d_array, (2, 1, 0))
        arr = [(i, j, k) for i in range(8)
               for j in range(8)
               for k in range(8)]
        self.assertTrue(is_sorted_array_of_nd_positions(arr, key, 8))

    def test_not_is_sorted_2d_last_elements_swapped(self):
        key = (index_of_position_in_1d_array, (1, 0))
        arr = [(i, j) for i in range(12) for j in range(12)]
        arr[-1], arr[-2] = arr[-2], arr[-1]
        self.assertFalse(is_sorted_array_of_nd_positions(arr, key, 12))


class ParallelBubbleSort1DTest(unittest.TestCase):
    def test_sort_4_elements(self):
        key = (index_of_position_in_1d_array, (2, 1, 0))
        arr = [(1, 0, 0), (0, 0, 1), (0, 1, 0), (0, 0, 0)]
        parallel_bubble_sort(arr, key, system_side_length=2)
        self.assertListEqual(arr,
                             [(0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0)])

    def test_sort_reverse_long_no_optional_parameters_specified(self):
        key = (index_of_position_in_1d_array, (2, 0, 1))
        arr = [(0, i, 0) for i in range(10, -1, -1)]
        parallel_bubble_sort(arr, key, system_side_length=11)
        self.assertListEqual(arr, [(0, i, 0) for i in range(11)])

    def test_sort_scrambled(self):
        key = (index_of_position_in_1d_array, (0, 1))
        arr = [(1, 2), (0, 1), (2, 1), (1, 0), (0, 0), (2, 0)]
        parallel_bubble_sort(arr, key, system_side_length=3)
        self.assertListEqual(arr, [(0, 0), (1, 0), (2, 0),
                                   (0, 1), (2, 1), (1, 2)])

    def test_sort_already_sorted(self):
        key = (index_of_position_in_1d_array, (2, 1, 0))
        arr = [(0, i, j) for i in range(5) for j in range(5)]
        parallel_bubble_sort(arr, key, system_side_length=5)
        self.assertListEqual(arr, [(0, i, j) for i in range(5)
                                   for j in range(5)])


class ParallelBubbleSortIntegrationTest(unittest.TestCase):
    def test_sort_2d_grid_embedded_in_1d_array_integration_test(self):
        from functools import partial

        key_2d = partial(index_of_position_in_1d_array, (0, 1))
        system_size = 8
        arr = [(i, j) for i in range(system_size)
               for j in range(system_size)]

        self.assertListEqual(list(map(partial(key_2d, system_size), arr))[:16],
                             list(range(0, 64, 8)) + list(range(1, 65, 8)))

        self.assertFalse(is_sorted_array_of_nd_positions(arr, key_2d,
                                                         system_size))
        swaps = parallel_bubble_sort(arr, key_2d, system_size)
        self.assertEqual(len(swaps), 50)
        self.assertTrue(is_sorted_array_of_nd_positions(arr, key_2d,
                                                        system_size))

    def test_ffft_like_sorting(self):
        # Determine how to order the modes for correct input to the FFFT.
        system_size = 8
        s = '{:0' + str(3) + 'b}'
        first_round = [int(s.format(i)[::-1], 2) for i in range(system_size)]
        second_round = (list(range(system_size // 2, system_size)) +
                        list(range(system_size // 2)))

        # Generate array of size system_size.
        arr = [(i,) for i in range(system_size)]

        # Create arrays whose swap networks to resort are the same as should
        # be applied to the fermionic modes.
        arr_first = [None] * system_size
        arr_second = [None] * system_size
        for i in range(system_size):
            arr_first[i] = (first_round[arr[i][0]],)
            arr_second[i] = (second_round[arr[i][0]],)

        # Find the two rounds of swaps for parallel bubble sort into
        # row-major order.
        key = (index_of_position_in_1d_array, (0,))
        first_round_swaps = parallel_bubble_sort(arr_first, key, system_size)
        second_round_swaps = parallel_bubble_sort(arr_second, key, system_size)

        # Create the full list of swaps.
        swap_mode_list = first_round_swaps + second_round_swaps[::-1]
        swaps = [swap[0] for swap_layer in swap_mode_list
                 for swap in swap_layer]

        # Compare with the correct full swap network for the FFFT.
        self.assertListEqual(swaps, [1, 3, 5, 2, 4, 1, 3, 5, 3, 2, 4,
                                     1, 3, 5, 0, 2, 4, 6, 1, 3, 5, 2, 4, 3])



"""Tests for _low_depth_trotter_simulation.py."""
import numpy
import os
import random
import unittest

import projectq

from scipy.sparse.linalg import expm

from openfermionprojectq import _low_depth_trotter_simulation

from openfermion.hamiltonians import plane_wave_hamiltonian
from openfermion.ops import FermionOperator, normal_ordered
from openfermion.transforms import get_sparse_operator
from openfermion.utils import Grid


os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


def ordered_wavefunction(engine, indices_to_evaluate=None):
    """Return the correctly ordered wave function amplitudes.

    Args:
        engine (projectq.MainEngine): The engine from which to take the
                                      amplitudes.
        indices_to_evaluate (numpy.array): Array of integer indices to
                                           evaluate amplitudes of. If
                                           indices_to_evaluate is not
                                           set, all amplitudes are
                                           evaluated.
    """
    n_qubits = engine._qubit_idx
    if indices_to_evaluate is None:
        indices_to_evaluate = numpy.arange(2 ** n_qubits)

    indices = numpy.zeros(len(indices_to_evaluate), dtype=int)

    # Get the qubit order dictionary and raw wavefunction from the engine.
    qubit_order_dict, unordered_wavefunction = engine.backend.cheat()

    for bit_number in range(n_qubits):
        # To each logical state for which qubit k is 1, add
        # 2 ** (qubit k's position in the raw wavefunction). This is quickly
        # given by 1 << qubit_order_dict[k].
        indices[(indices_to_evaluate & (1 << bit_number)) != 0] += (
            1 << qubit_order_dict[bit_number])

    # Return the subset of unordered_wavefunction corresponding to the indices.
    return list(map(unordered_wavefunction.__getitem__, indices))


def dual_basis_hamiltonian(n_dimensions, system_size,
                           plane_wave=False, spinless=True):
    """Return the plane wave dual basis hamiltonian with the given parameters.

    Returns: the Hamiltonian as a normal-ordered FermionOperator.

    Args:
        n_dimensions (int): The number of dimensions in the system.
        system_size (int): The side length along each dimension.
        plane_wave (bool): Whether the system is in the plane wave basis.
        spinless (bool): Whether the system is spinless.

    Notes:
        Follows Eq. 10 of https://arxiv.org/pdf/1706.00023.pdf.
    """
    grid = Grid(n_dimensions, length=system_size, scale=1.0 * system_size)
    hamiltonian = plane_wave_hamiltonian(grid, spinless=spinless,
                                         plane_wave=plane_wave)
    return normal_ordered(hamiltonian)


class FourQubitSecondOrderTrotterTest(unittest.TestCase):

    def setUp(self):
        self.size = 4
        self.engine = projectq.MainEngine()
        self.register = self.engine.allocate_qureg(self.size)
        random.seed(17)

    def tearDown(self):
        projectq.ops.All(projectq.ops.Measure) | self.register

    def test_simulate_n0n1(self):
        hamiltonian = FermionOperator('1^ 0^ 1 0')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Use 3^ 2^ 3 2 because get_sparse_operator reverses the indices.
        evol_matrix = expm(-1j * get_sparse_operator(
            FermionOperator('3^ 2^ 3 2'), n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(ordered_wavefunction(self.engine) -
                                expected.T))

    def test_simulate_n0n3(self):
        hamiltonian = FermionOperator('3^ 0^ 3 0')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Though get_sparse_operator reverses the indices, this is symmetric.
        evol_matrix = expm(-1j * get_sparse_operator(
            hamiltonian, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(ordered_wavefunction(self.engine) -
                                expected.T))

    def test_simulate_n1n3(self):
        hamiltonian = FermionOperator('3^ 1^ 3 1')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Use 2^ 0^ 2 0 because get_sparse_operator reverses the indices.
        evol_matrix = expm(-1j * get_sparse_operator(
            FermionOperator('2^ 0^ 2 0'), n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(ordered_wavefunction(self.engine) -
                                expected.T))

    def test_single_trotter_step_no_input_ordering_n1n3(self):
        hamiltonian = FermionOperator('3^ 1^ 3 1')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulation_gate_trotter_step(
            self.register, hamiltonian, first_order=False)
        self.engine.flush()

        # Use 2^ 0^ 2 0 because get_sparse_operator reverses the indices.
        evol_matrix = expm(-1j * get_sparse_operator(
            FermionOperator('2^ 0^ 2 0'), n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(ordered_wavefunction(self.engine) -
                                expected.T))

    def test_simulate_hopping_0_to_1(self):
        hamiltonian = FermionOperator('1^ 0') + FermionOperator('0^ 1')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Use 3^ 2 + 2^ 3 because get_sparse_operator reverses the indices.
        reversed_operator = FermionOperator('3^ 2') + FermionOperator('2^ 3')
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(ordered_wavefunction(self.engine) -
                                expected.T))

    def test_simulate_hopping_1_to_3(self):
        hamiltonian = FermionOperator('1^ 3') + FermionOperator('3^ 1')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Use 0^ 2 + 2^ 0 because get_sparse_operator reverses the indices.
        reversed_operator = FermionOperator('0^ 2') + FermionOperator('2^ 0')
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T))

    def test_simulate_hopping_0_to_3(self):
        hamiltonian = FermionOperator('0^ 3') + FermionOperator('3^ 0')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)
        self.engine.flush()

        # Use 0^ 2 + 2^ 0 because get_sparse_operator reverses the indices.
        evol_matrix = expm(-1j * get_sparse_operator(
            hamiltonian, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T))

    def test_simulate_multiple_hopping_terms(self):
        hamiltonian = (FermionOperator('0^ 3') + FermionOperator('3^ 0') +
                       2 * (FermionOperator('1^ 3') +
                            FermionOperator('3^ 1')) -
                       0.5 * (FermionOperator('0^ 1') +
                              FermionOperator('1^ 0')))

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=4, first_order=False)

        self.engine.flush()

        # get_sparse_operator reverses the indices, so reverse the sites
        # the Hamiltonian acts on so as to compare them.
        reversed_operator = (
            FermionOperator('0^ 3') + FermionOperator('3^ 0') +
            2 * (FermionOperator('0^ 2') + FermionOperator('2^ 0')) -
            0.5 * (FermionOperator('3^ 2') + FermionOperator('2^ 3')))
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T, atol=1e-2))

    def test_simulate_multiple_two_number_terms(self):
        hamiltonian = (0.37 * FermionOperator('1^ 0^ 1 0') +
                       2.4 * FermionOperator('3^ 0^ 3 0') -
                       2 * FermionOperator('3^ 1^ 3 1'))

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)

        self.engine.flush()

        # get_sparse_operator reverses the indices, so reverse the sites
        # the Hamiltonian acts on so as to compare them.
        reversed_operator = (0.37 * FermionOperator('3^ 2^ 3 2') +
                             2.4 * FermionOperator('3^ 0^ 3 0') -
                             2 * FermionOperator('2^ 0^ 2 0'))
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(numpy.array(ordered_wavefunction(self.engine) -
                                            expected.T)))

    def test_simulate_single_and_double_number_terms(self):
        hamiltonian = (0.37 * FermionOperator('1^ 0^ 1 0') +
                       2.4 * FermionOperator('3^ 0^ 3 0') -
                       2 * FermionOperator('3^ 1^ 3 1') +
                       1.1 * FermionOperator('2^ 2') +
                       1.7 * FermionOperator('0^ 0') -
                       0.3 * FermionOperator('3^ 3'))

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=1, first_order=False)

        self.engine.flush()

        # get_sparse_operator reverses the indices, so reverse the sites
        # the Hamiltonian acts on so as to compare them.
        reversed_operator = (0.37 * FermionOperator('3^ 2^ 3 2') +
                             2.4 * FermionOperator('3^ 0^ 3 0') -
                             2 * FermionOperator('2^ 0^ 2 0') +
                             1.1 * FermionOperator('1^ 1') +
                             1.7 * FermionOperator('3^ 3') -
                             0.3 * FermionOperator('0^ 0'))
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T),
                        msg=str(numpy.array(ordered_wavefunction(self.engine) -
                                            expected.T)))

    def test_simulate_overlapping_number_and_hopping_terms(self):
        hamiltonian = (0.37 * FermionOperator('1^ 0^ 1 0') +
                       FermionOperator('2^ 0') + FermionOperator('0^ 2'))

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=2, first_order=False)

        self.engine.flush()

        # get_sparse_operator reverses the indices, so reverse the sites
        # the Hamiltonian acts on so as to compare them.
        reversed_operator = (0.37 * FermionOperator('3^ 2^ 3 2') +
                             FermionOperator('3^ 1') + FermionOperator('1^ 3'))
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T,
                                       atol=1e-2),
                        msg=str(numpy.array(ordered_wavefunction(self.engine) -
                                            expected.T)))

    def test_simulate_dual_basis_hamiltonian(self):
        hamiltonian = dual_basis_hamiltonian(1, self.size)
        self.engine.flush()

        # Choose random state.
        initial_state = numpy.zeros(2 ** self.size, dtype=complex)
        for i in range(len(initial_state)):
            initial_state[i] = (random.random() *
                                numpy.exp(1j * 2 * numpy.pi * random.random()))
        initial_state /= numpy.linalg.norm(initial_state)

        # Put randomly chosen state in the registers.
        self.engine.backend.set_wavefunction(initial_state, self.register)

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=7, first_order=False)

        self.engine.flush()

        # get_sparse_operator reverses the indices, but the jellium
        # Hamiltonian is symmetric.
        evol_matrix = expm(-1j * get_sparse_operator(
            hamiltonian, n_qubits=self.size)).todense()
        initial_state = numpy.matrix(initial_state).T
        expected = evol_matrix * initial_state

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T,
                                       atol=1e-2),
                        msg=str(numpy.array(ordered_wavefunction(self.engine) -
                                            expected.T)))

    def test_simulate_dual_basis_hamiltonian_with_spin_and_potentials(self):
        big_eng = projectq.MainEngine()
        big_reg = big_eng.allocate_qureg(2 * self.size)
        hamiltonian = dual_basis_hamiltonian(1, self.size, spinless=False)

        for i in range(2 * self.size):
            coefficient = 1. / (i + 1)
            if i % 3:
                coefficient = -coefficient
            hamiltonian += FermionOperator(((i, 1), (i, 0)), coefficient)

        # Choose random state.
        initial_state = numpy.zeros(2 ** (2 * self.size), dtype=complex)
        for i in range(len(initial_state)):
            initial_state[i] = (random.random() *
                                numpy.exp(1j * 2 * numpy.pi * random.random()))
        initial_state /= numpy.linalg.norm(initial_state)

        # Put randomly chosen state in the registers.
        big_eng.flush()
        big_eng.backend.set_wavefunction(initial_state, big_reg)

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            big_reg, hamiltonian, trotter_steps=7, first_order=False,
            input_ordering=list(range(7, -1, -1)))

        big_eng.flush()

        # get_sparse_operator reverses the indices - we've accounted for this
        # with the reversed input_ordering.
        evol_matrix = expm(-1j * get_sparse_operator(
            hamiltonian, n_qubits=2*self.size)).todense()
        initial_state = numpy.matrix(initial_state).T
        expected = evol_matrix * initial_state

        self.assertTrue(numpy.allclose(ordered_wavefunction(big_eng),
                                       expected.T,
                                       atol=1e-2),
                        msg=str(numpy.array(ordered_wavefunction(big_eng) -
                                            expected.T)))

        projectq.ops.All(projectq.ops.Measure) | big_reg

    def test_simulate_dual_basis_evolution_bad_input_ordering(self):
        with self.assertRaises(ValueError):
            _low_depth_trotter_simulation.simulate_dual_basis_evolution(
                self.register, FermionOperator(), input_ordering=[1, 2])

    def test_simulate_dual_basis_evolution_n_trotter_steps_not_integer(self):
        with self.assertRaises(ValueError):
            _low_depth_trotter_simulation.simulate_dual_basis_evolution(
                self.register, FermionOperator(), trotter_steps=1.5)

    def test_simulate_dual_basis_evolution_bad_n_trotter_steps(self):
        with self.assertRaises(ValueError):
            _low_depth_trotter_simulation.simulate_dual_basis_evolution(
                self.register, FermionOperator(), trotter_steps=0)


class FourQubitFirstOrderEquivalenceWithSecondOrderTest(unittest.TestCase):

    def setUp(self):
        self.size = 4
        self.engine = projectq.MainEngine()
        self.register = self.engine.allocate_qureg(self.size)

    def tearDown(self):
        projectq.ops.All(projectq.ops.Measure) | self.register

    def test_first_order_odd_number_of_steps_reversal(self):
        hamiltonian = FermionOperator('1^ 3') + FermionOperator('3^ 1')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=3, first_order=True)

        _low_depth_trotter_simulation.fermionic_reorder(
            self.register, range(self.size - 1, -1, -1))
        self.engine.flush()

        # Use 0^ 2 + 2^ 0 because get_sparse_operator reverses the indices.
        reversed_operator = FermionOperator('0^ 2') + FermionOperator('2^ 0')
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T))

    def test_first_order_even_number_of_steps_no_reversal(self):
        hamiltonian = FermionOperator('2^ 3') + FermionOperator('3^ 2')

        projectq.ops.All(projectq.ops.H) | self.register

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            self.register, hamiltonian, trotter_steps=2, first_order=True)

        self.engine.flush()

        # Use 0^ 1 + 1^ 0 because get_sparse_operator reverses the indices.
        reversed_operator = FermionOperator('0^ 1') + FermionOperator('1^ 0')
        evol_matrix = expm(-1j * get_sparse_operator(
            reversed_operator, n_qubits=self.size)).todense()
        expected = evol_matrix * numpy.matrix([2 ** (-self.size / 2.)] *
                                              2 ** self.size).T

        self.assertTrue(numpy.allclose(ordered_wavefunction(self.engine),
                                       expected.T))


class HighTrotterNumberIntegrationTest(unittest.TestCase):

    def setUp(self):
        random.seed(17)

    def test_trotter_order_does_not_matter_for_high_trotter_number(self):
        size = 4
        hamiltonian = dual_basis_hamiltonian(n_dimensions=1, system_size=size)
        hamiltonian.compress()

        eng1 = projectq.MainEngine()
        eng2 = projectq.MainEngine()
        reg1 = eng1.allocate_qureg(size)
        reg2 = eng2.allocate_qureg(size)
        eng1.flush()
        eng2.flush()

        # Choose random state.
        state = numpy.zeros(2 ** size, dtype=complex)
        for i in range(len(state)):
            state[i] = (random.random() *
                        numpy.exp(1j * 2 * numpy.pi * random.random()))
        state /= numpy.linalg.norm(state)

        # Put randomly chosen state in the registers.
        eng1.backend.set_wavefunction(state, reg1)
        eng2.backend.set_wavefunction(state, reg2)

        # Swap and change the input ordering for reg1. These operations should
        # cancel each other out.
        projectq.ops.Swap | (reg1[1], reg1[2])
        projectq.ops.C(projectq.ops.Z) | (reg1[1], reg1[2])

        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            reg1, hamiltonian=hamiltonian, trotter_steps=7,
            input_ordering=[0, 2, 1, 3], first_order=False)
        _low_depth_trotter_simulation.simulate_dual_basis_evolution(
            reg2, hamiltonian=hamiltonian, trotter_steps=7, first_order=False)

        # Undo the inital swaps on reg1.
        projectq.ops.Swap | (reg1[1], reg1[2])
        projectq.ops.C(projectq.ops.Z) | (reg1[1], reg1[2])

        projectq.ops.All(projectq.ops.H) | reg1
        eng1.flush()
        projectq.ops.All(projectq.ops.H) | reg2
        eng2.flush()

        self.assertTrue(numpy.allclose(ordered_wavefunction(eng1),
                                       ordered_wavefunction(eng2),
                                       atol=1e-2))
        projectq.ops.All(projectq.ops.Measure) | reg1
        projectq.ops.All(projectq.ops.Measure) | reg2



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import copy

import projectq

from openfermion.ops import QubitOperator


def TimeEvolution(time, hamiltonian):
    """
    Converts the Hamiltonian to an instance of the ProjectQ QubitOperator
    and then returns an instance of the ProjectQ TimeEvolution gate.
    This gate is the unitary time evolution propagator:
    exp(-i * H * t),
    where H is the Hamiltonian of the system and t is the time. Note that -i
    factor is stored implicitly.

    Example:
            wavefuction = eng.allocate_qureg(5)
            hamiltonian = 0.5 * QubitOperator("X0 Z1 Y5")
            # Apply exp(-i * H * t) to the wavefunction:
            TimeEvolution(time=2.0, hamiltonian=hamiltonian) | wavefunction

    Args:
        time(float, int): time t
        hamiltonian(QubitOperator): hamiltonaian H

    Returns:
        Instance of ProjectQ TimeEvolution gate.

    Raises:
        TypeError: If time is not a numeric type and hamiltonian is not a
            QubitOperator.
        NotHermitianOperatorError: If the input hamiltonian is not
            hermitian (only real coefficients).
    """
    projectq_qubit_operator = projectq.ops.QubitOperator()
    for term, coefficient in hamiltonian.terms.items():
        projectq_qubit_operator.terms[term] = coefficient
    return projectq.ops.TimeEvolution(time, projectq_qubit_operator)
